#    Copyright 2017 ARM Holdings.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Websocket proxy that is compatible with OpenStack Zun.
Leverages websockify.py by Joel Martin
"""

import errno
import select
import socket
import sys
import time

import docker
from oslo_log import log as logging
from oslo_utils import uuidutils
from urllib import parse as urlparse
import websockify

from zun.common import context
from zun.common import exception
from zun.common.i18n import _
import zun.conf
from zun import objects
from zun.websocket.websocketclient import WebSocketClient

LOG = logging.getLogger(__name__)
CONF = zun.conf.CONF


class ZunProxyRequestHandlerBase(object):
    def verify_origin_proto(self, access_url, origin_proto):
        if not access_url:
            detail = _("No access_url available."
                       "Cannot validate protocol")
            raise exception.ValidationError(detail=detail)
        expected_protos = [urlparse.urlparse(access_url).scheme]
        # NOTE: For serial consoles the expected protocol could be ws or
        # wss which correspond to http and https respectively in terms of
        # security.
        if 'ws' in expected_protos:
            expected_protos.append('http')
        if 'wss' in expected_protos:
            expected_protos.append('https')

        return origin_proto in expected_protos

    def _send_buffer(self, buff, target, send_all=False):
        size = len(buff)
        tosend = size
        already_sent = 0

        while tosend > 0:
            try:
                # i should be able to send a bytearray
                sent = target.send(buff[already_sent:])
                if sent == 0:
                    raise RuntimeError(_('socket connection broken'))

                already_sent += sent
                tosend -= sent

            except socket.error as e:
                # if full buffers then wait for them to drain and try again
                if e.errno in [errno.EAGAIN, errno.EWOULDBLOCK]:
                    if send_all:
                        continue
                    return buff[already_sent:]
                else:
                    raise exception.SocketException(str(e))
        return None

    def _handle_ins_outs(self, target, ins, outs):
        """Handle the select file ins and outs

        Handle the operation ins and outs from select
        """
        if self.request in outs:
            # Send queued target data to the client
            self.c_pend = self.send_frames(self.cqueue)
            self.cqueue = []

        if self.request in ins:
            # Receive client data, decode it, and queue for target
            bufs, closed = self.recv_frames()
            self.tqueue.extend(bufs)
            if closed:
                self.msg(_("Client closed connection:"
                           "%(host)s:%(port)s") % {
                    'host': self.server.target_host,
                    'port': self.server.target_port})
                raise self.CClose(closed['code'], closed['reason'])

        if target in outs:
            while self.tqueue:
                payload = self.tqueue.pop(0)
                remaining = self._send_buffer(payload, target)
                if remaining is not None:
                    self.tqueue.appendleft(remaining)
                    break

        if target in ins:
            # Receive target data, encode it and queue for client
            buf = target.recv()
            if len(buf) == 0:
                self.msg(_("Client closed connection:"
                           "%(host)s:%(port)s") % {
                    'host': self.server.target_host,
                    'port': self.server.target_port})
                raise self.CClose(1000, "Target closed")
            if isinstance(buf, str):
                buf = buf.encode()
            self.cqueue.append(buf)

    def do_websocket_proxy(self, target):
        """Proxy websocket link

        Proxy client WebSocket to normal target socket.
        """
        self.cqueue = []
        self.tqueue = []
        self.c_pend = 0
        rlist = [self.request, target]

        if self.server.heartbeat:
            now = time.time()
            self.heartbeat = now + self.server.heartbeat
        else:
            self.heartbeat = None

        while True:
            wlist = []

            if self.heartbeat is not None:
                now = time.time()
                if now > self.heartbeat:
                    self.heartbeat = now + self.server.heartbeat
                    self.send_ping()

            if self.tqueue:
                wlist.append(target)
            if self.cqueue or self.c_pend:
                wlist.append(self.request)
            try:
                ins, outs, excepts = select.select(rlist, wlist, [], 1)
            except (select.error, OSError):
                exc = sys.exc_info()[1]
                if hasattr(exc, 'errno'):
                    err = exc.errno
                else:
                    err = exc[0]
                if err != errno.EINTR:
                    raise
                else:
                    continue

            if excepts:
                raise exception.SocketException()

            self._handle_ins_outs(target, ins, outs)

    def new_websocket_client(self):
        """Called after a new WebSocket connection has been established."""
        # Reopen the eventlet hub to make sure we don't share an epoll
        # fd with parent and/or siblings, which would be bad
        from eventlet import hubs
        hubs.use_hub()

        # The zun expected behavior is to have token
        # passed to the method GET of the request
        parse = urlparse.urlparse(self.path)
        if parse.scheme not in ('http', 'https'):
            # From a bug in urlparse in Python < 2.7.4 we cannot support
            # special schemes (cf: https://bugs.python.org/issue9374)
            if sys.version_info < (2, 7, 4):
                raise exception.ZunException(
                    _("We do not support scheme '%s' under Python < 2.7.4, "
                      "please use http or https") % parse.scheme)

        query = parse.query
        token = urlparse.parse_qs(query).get("token", [""]).pop()
        uuid = urlparse.parse_qs(query).get("uuid", [""]).pop()
        exec_id = urlparse.parse_qs(query).get("exec_id", [""]).pop()

        ctx = context.get_admin_context(all_projects=True)

        if uuidutils.is_uuid_like(uuid):
            container = objects.Container.get_by_uuid(ctx, uuid)
        else:
            container = objects.Container.get_by_name(ctx, uuid)

        if exec_id:
            self._new_exec_client(container, token, uuid, exec_id)
        else:
            self._new_websocket_client(container, token, uuid)

    def _new_websocket_client(self, container, token, uuid):
        if token != container.websocket_token:
            raise exception.InvalidWebsocketToken(token)

        access_url = '%s?token=%s&uuid=%s' % (CONF.websocket_proxy.base_url,
                                              token, uuid)

        self._verify_origin(access_url)

        if container.websocket_url:
            target_url = container.websocket_url
            escape = "~"
            close_wait = 0.5
            wscls = WebSocketClient(host_url=target_url, escape=escape,
                                    close_wait=close_wait)
            wscls.connect()
            self.target = wscls
        else:
            raise exception.InvalidWebsocketUrl()

        # Start proxying
        try:
            self.do_websocket_proxy(self.target.ws)
        except Exception:
            if self.target.ws:
                self.target.ws.close()
                self.vmsg(_("Websocket client or target closed"))
            raise

    def _new_exec_client(self, container, token, uuid, exec_id):
        exec_instance = None
        for e in container.exec_instances:
            if token == e.token and exec_id == e.exec_id:
                exec_instance = e

        if not exec_instance:
            raise exception.InvalidWebsocketToken(token)

        access_url = '%s?token=%s&uuid=%s' % (CONF.websocket_proxy.base_url,
                                              token, uuid)

        self._verify_origin(access_url)

        client = docker.APIClient(base_url=exec_instance.url)
        tsock = client.exec_start(exec_id, socket=True, tty=True)
        if hasattr(tsock, "_sock"):
            # NOTE(hongbin): dockerpy returns different socket class depending
            # on python version and base_url (see _get_raw_response_socket) so
            # we need to handle it in here.
            tsock = tsock._sock

        try:
            self.do_proxy(tsock)
        finally:
            if tsock:
                tsock.shutdown(socket.SHUT_RDWR)
                tsock.close()
                self.vmsg(_("%s: Closed target") % exec_instance.url)

    def _verify_origin(self, access_url):
        # Verify Origin
        expected_origin_hostname = self.headers.get('Host')
        if ':' in expected_origin_hostname:
            e = expected_origin_hostname
            if '[' in e and ']' in e:
                expected_origin_hostname = e.split(']')[0][1:]
            else:
                expected_origin_hostname = e.split(':')[0]
        expected_origin_hostnames = CONF.websocket_proxy.allowed_origins
        expected_origin_hostnames.append(expected_origin_hostname)
        origin_url = self.headers.get('Origin')

        # missing origin header indicates non-browser client which is OK
        if origin_url is not None:
            origin = urlparse.urlparse(origin_url)
            origin_hostname = origin.hostname
            origin_scheme = origin.scheme
            # If the console connection was forwarded by a proxy (example:
            # haproxy), the original protocol could be contained in the
            # X-Forwarded-Proto header instead of the Origin header. Prefer the
            # forwarded protocol if it is present.
            forwarded_proto = self.headers.get('X-Forwarded-Proto')
            if forwarded_proto is not None:
                origin_scheme = forwarded_proto
            if origin_hostname == '' or origin_scheme == '':
                detail = _("Origin header not valid.")
                raise exception.ValidationError(detail)
            if origin_hostname not in expected_origin_hostnames:
                detail = _("Origin header does not match this host.")
                raise exception.ValidationError(detail)
            if not self.verify_origin_proto(access_url, origin_scheme):
                detail = _("Origin header protocol does not match this host.")
                raise exception.ValidationError(detail)


class ZunProxyRequestHandler(ZunProxyRequestHandlerBase,
                             websockify.ProxyRequestHandler):
    def __init__(self, *args, **kwargs):
        websockify.ProxyRequestHandler.__init__(self, *args, **kwargs)

    def socket(self, *args, **kwargs):
        return websockify.WebSocketServer.socket(*args, **kwargs)


class ZunWebSocketProxy(websockify.WebSocketProxy):
    @staticmethod
    def get_logger():
        return LOG



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import strutils
import pecan

from zun.api.controllers import base
from zun.api.controllers.v1 import collection
from zun.api.controllers.v1.schemas import services as schema
from zun.api.controllers.v1.views import services_view as view
from zun.api import servicegroup as svcgrp_api
from zun.api import validation
from zun.common import exception
from zun.common import policy
import zun.conf
from zun import objects


CONF = zun.conf.CONF


class ZunServiceCollection(collection.Collection):

    fields = {
        'services',
        'next'
    }

    def __init__(self, **kwargs):
        super(ZunServiceCollection, self).__init__()
        self._type = 'services'

    @staticmethod
    def convert_db_rec_list_to_collection(servicegroup_api,
                                          rpc_hsvcs, **kwargs):
        collection = ZunServiceCollection()
        collection.services = []
        for p in rpc_hsvcs:
            service = p.as_dict()
            alive = servicegroup_api.service_is_up(p)
            state = 'up' if alive else 'down'
            service['state'] = state
            service = view.format_service(service)
            collection.services.append(service)
            if not service['availability_zone']:
                service['availability_zone'] = CONF.default_availability_zone
        next = collection.get_next(limit=None, url=None, **kwargs)
        if next is not None:
            collection.next = next
        return collection


class ZunServiceController(base.Controller):
    """REST controller for zun-services."""

    _custom_actions = {
        'enable': ['PUT'],
        'disable': ['PUT'],
        'force_down': ['PUT'],
    }

    def __init__(self, **kwargs):
        super(ZunServiceController, self).__init__()
        self.servicegroup_api = svcgrp_api.ServiceGroup()

    def _update(self, context, host, binary, payload):
        """Do the actual update"""
        svc = objects.ZunService.get_by_host_and_binary(
            context, host, binary)
        if svc is None:
            raise exception.ZunServiceNotFound(
                binary=binary, host=host)
        else:
            return svc.update(context, payload)

    def _enable_or_disable(self, context, body, params_to_update):
        """Enable/Disable scheduling for a service."""
        self._update(context, body['host'], body['binary'],
                     params_to_update)
        res = {
            'service': {
                'host': body['host'],
                'binary': body['binary'],
                'disabled': params_to_update['disabled'],
                'disabled_reason': params_to_update['disabled_reason']
            },
        }
        return res

    def _enable(self, context, body):
        """Enable scheduling for a service."""
        return self._enable_or_disable(context, body,
                                       {'disabled': False,
                                        'disabled_reason': None})

    def _disable(self, context, body, reason=None):
        """Disable scheduling for a service with optional log."""
        return self._enable_or_disable(context, body,
                                       {'disabled': True,
                                        'disabled_reason': reason})

    def _update_forced_down(self, context, body):
        """Set or unset forced_down flag for the service"""
        try:
            forced_down = strutils.bool_from_string(body['forced_down'], True)
        except ValueError:
            bools = ', '.join(strutils.TRUE_STRINGS + strutils.FALSE_STRINGS)
            raise exception.InvalidValue(_('Valid forced_down values are: %s')
                                         % bools)
        self._update(context, body['host'], body['binary'],
                     {"forced_down": forced_down})
        res = {
            'service': {
                'host': body['host'],
                'binary': body['binary'],
                'forced_down': forced_down
            },
        }
        return res

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_all(self, **kwargs):
        """Retrieve a list of zun-services.

        """
        context = pecan.request.context
        policy.enforce(context, "zun-service:get_all",
                       action="zun-service:get_all")
        hsvcs = objects.ZunService.list(context,
                                        limit=None,
                                        marker=None,
                                        sort_key='id',
                                        sort_dir='asc')
        return ZunServiceCollection.convert_db_rec_list_to_collection(
            self.servicegroup_api, hsvcs)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def delete(self, host, binary):
        """Delete the specified service.

        :param host: The host on which the binary is running.
        :param binary: The name of the binary.
        """
        context = pecan.request.context
        policy.enforce(context, "zun-service:delete",
                       action="zun-service:delete")
        svc = objects.ZunService.get_by_host_and_binary(
            context, host, binary)
        if svc is None:
            raise exception.ZunServiceNotFound(
                binary=binary, host=host)
        else:
            svc.destroy(context)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    @validation.validate_query_param(pecan.request,
                                     schema.query_param_enable)
    def enable(self, **kwargs):
        context = pecan.request.context
        policy.enforce(context, "zun-service:enable",
                       action="zun-service:enable")
        return self._enable(context, kwargs)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    @validation.validate_query_param(pecan.request,
                                     schema.query_param_disable)
    def disable(self, **kwargs):
        context = pecan.request.context
        policy.enforce(context, "zun-service:disable",
                       action="zun-service:disable")
        if 'disabled_reason' in kwargs:
            reason = kwargs['disabled_reason']
        else:
            reason = None
        return self._disable(context, kwargs, reason)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    @validation.validate_query_param(pecan.request,
                                     schema.query_param_force_down)
    def force_down(self, **kwargs):
        context = pecan.request.context
        policy.enforce(context, "zun-service:force_down",
                       action="zun-service:force_down")
        return self._update_forced_down(context, kwargs)



# Copyright 2016 IBM, Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_concurrency import lockutils


class Singleton(type):
    _instances = {}
    _semaphores = lockutils.Semaphores()

    def __call__(cls, *args, **kwargs):
        with lockutils.lock('singleton_lock', semaphores=cls._semaphores):
            if cls not in cls._instances:
                cls._instances[cls] = super(
                    Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cinderclient import exceptions as cinder_exception
from oslo_utils import timeutils

import zun.conf
from zun.tests import base
from zun.volume import cinder_api


CONF = zun.conf.CONF


class FakeVolume(object):

    def __init__(self, volume_id, size=1, attachments=None, multiattach=False):
        self.id = volume_id
        self.name = 'volume_name'
        self.description = 'volume_description'
        self.status = 'available'
        self.created_at = timeutils.utcnow()
        self.size = size
        self.availability_zone = 'nova'
        self.attachments = attachments or []
        self.volume_type = 99
        self.bootable = False
        self.snapshot_id = 'snap_id_1'
        self.metadata = {}
        self.multiattach = multiattach


class TestingException(Exception):
    pass


class CinderApiTestCase(base.TestCase):

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_get(self, mock_cinderclient):
        volume_id = 'volume_id1'
        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)

        self.api = cinder_api.CinderAPI(self.context)
        self.api.get(volume_id)

        mock_cinderclient.assert_called_once_with()
        mock_volumes.get.assert_called_once_with(volume_id)

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_reserve_volume(self, mock_cinderclient):
        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)

        self.api = cinder_api.CinderAPI(self.context)
        self.api.reserve_volume('id1')

        mock_cinderclient.assert_called_once_with()
        mock_volumes.reserve.assert_called_once_with('id1')

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_unreserve_volume(self, mock_cinderclient):
        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)

        self.api = cinder_api.CinderAPI(self.context)
        self.api.unreserve_volume('id1')

        mock_cinderclient.assert_called_once_with()
        mock_volumes.unreserve.assert_called_once_with('id1')

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_begin_detaching(self, mock_cinderclient):
        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)

        self.api = cinder_api.CinderAPI(self.context)
        self.api.begin_detaching('id1')

        mock_cinderclient.assert_called_once_with()
        mock_volumes.begin_detaching.assert_called_once_with('id1')

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_roll_detaching(self, mock_cinderclient):
        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)

        self.api = cinder_api.CinderAPI(self.context)
        self.api.roll_detaching('id1')

        mock_cinderclient.assert_called_once_with()
        mock_volumes.roll_detaching.assert_called_once_with('id1')

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_attach(self, mock_cinderclient):
        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)

        self.api = cinder_api.CinderAPI(self.context)
        self.api.attach('id1', 'point', 'host')

        mock_cinderclient.assert_called_once_with()
        mock_volumes.attach.assert_called_once_with(
            volume='id1', mountpoint='point', host_name='host',
            instance_uuid=None)

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_detach(self, mock_cinderclient):
        attachment = {'server_id': 'fake_server',
                      'attachment_id': 'fakeid'}

        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)
        mock_cinderclient.return_value.volumes.get.return_value = \
            FakeVolume('id1', attachments=[attachment])
        volume = mock.MagicMock()
        volume.cinder_volume_id = 'id1'

        self.api = cinder_api.CinderAPI(self.context)
        self.api.detach(volume)

        mock_cinderclient.assert_called_with()
        mock_volumes.detach.assert_called_once_with('id1', None)

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_detach_multiattach(self, mock_cinderclient):
        attachment = {'server_id': 'fake_server_id',
                      'attachment_id': 'fakeid'}

        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)
        mock_cinderclient.return_value.volumes.get.return_value = \
            FakeVolume('id1', attachments=[attachment], multiattach=True)
        volume = mock.MagicMock()
        volume.cinder_volume_id = 'id1'
        volume.container_uuid = 'fake_server_id'

        self.api = cinder_api.CinderAPI(self.context)
        self.api.detach(volume)

        mock_cinderclient.assert_called_with()
        mock_volumes.detach.assert_called_once_with('id1', 'fakeid')

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_initialize_connection(self, mock_cinderclient):
        connection_info = {'foo': 'bar'}
        mock_cinderclient.return_value.volumes. \
            initialize_connection.return_value = connection_info

        volume_id = 'fake_vid'
        connector = {'host': 'fakehost1'}
        self.api = cinder_api.CinderAPI(self.context)
        actual = self.api.initialize_connection(volume_id, connector)

        expected = connection_info
        self.assertEqual(expected, actual)

        mock_cinderclient.return_value.volumes. \
            initialize_connection.assert_called_once_with(volume_id, connector)

    @mock.patch('zun.volume.cinder_api.LOG')
    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_initialize_connection_exception_no_code(
            self, mock_cinderclient, mock_log):
        mock_cinderclient.return_value.volumes. \
            initialize_connection.side_effect = (
                cinder_exception.ClientException(500, "500"))
        mock_cinderclient.return_value.volumes. \
            terminate_connection.side_effect = (TestingException)

        connector = {'host': 'fakehost1'}
        self.api = cinder_api.CinderAPI(self.context)
        self.assertRaises(cinder_exception.ClientException,
                          self.api.initialize_connection,
                          'id1',
                          connector)
        self.assertIsNone(mock_log.error.call_args_list[1][0][1]['code'])

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_initialize_connection_rollback(self, mock_cinderclient):
        mock_cinderclient.return_value.volumes.\
            initialize_connection.side_effect = (
                cinder_exception.ClientException(500, "500"))

        connector = {'host': 'host1'}
        self.api = cinder_api.CinderAPI(self.context)
        ex = self.assertRaises(cinder_exception.ClientException,
                               self.api.initialize_connection,
                               'id1',
                               connector)
        self.assertEqual(500, ex.code)
        mock_cinderclient.return_value.volumes.\
            terminate_connection.assert_called_once_with('id1', connector)

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_initialize_connection_no_rollback(self, mock_cinderclient):
        mock_cinderclient.return_value.volumes.\
            initialize_connection.side_effect = TestingException

        connector = {'host': 'host1'}
        self.api = cinder_api.CinderAPI(self.context)
        self.assertRaises(TestingException,
                          self.api.initialize_connection,
                          'id1',
                          connector)
        self.assertFalse(mock_cinderclient.return_value.volumes.
                         terminate_connection.called)

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_terminate_connection(self, mock_cinderclient):
        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)

        self.api = cinder_api.CinderAPI(self.context)
        self.api.terminate_connection('id1', 'connector')

        mock_cinderclient.assert_called_once_with()
        mock_volumes.terminate_connection.assert_called_once_with('id1',
                                                                  'connector')

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_create_volume(self, mock_cinderclient):
        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)

        volume_size = '5'
        self.api = cinder_api.CinderAPI(self.context)
        self.api.create_volume(volume_size)

        mock_cinderclient.assert_called_once_with()
        mock_volumes.create.assert_called_once_with(volume_size)

    @mock.patch('zun.common.clients.OpenStackClients.cinder')
    def test_delete_volume(self, mock_cinderclient):
        mock_volumes = mock.MagicMock()
        mock_cinderclient.return_value = mock.MagicMock(volumes=mock_volumes)

        volume_id = self.id
        self.api = cinder_api.CinderAPI(self.context)
        self.api.delete_volume(volume_id)

        mock_cinderclient.assert_called_once_with()
        mock_volumes.delete.assert_called_once_with(volume_id)



#    Copyright 2017 ARM Holdings.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from zun.api.controllers import link
from zun.api.controllers.v1.views import containers_view
from zun.common.policies import capsule as policies


_basic_keys = (
    'uuid',
    'user_id',
    'project_id',
    'created_at',
    'updated_at',
    'addresses',
    'status',
    'status_reason',
    'restart_policy',
    'name',
    'labels',
    'annotations',
    'memory',
    'cpu',
    'init_containers',
    'containers',
    'host',
)


def format_capsule(url, capsule, context, legacy_api_version=False):
    def transform(key, value):
        if key not in _basic_keys:
            return
        # strip the key if it is not allowed by policy
        policy_action = policies.CAPSULE % ('get:%s' % key)
        if not context.can(policy_action, fatal=False, might_not_exist=True):
            return
        if key == 'uuid':
            yield ('uuid', value)
            yield ('links', [link.make_link(
                'self', url, 'capsules', value),
                link.make_link(
                    'bookmark', url,
                    'capsules', value,
                    bookmark=True)])
            if legacy_api_version:
                yield('volumes_info', {})
                yield('containers_uuids', [])
                yield('init_containers_uuids', [])
                yield('capsule_version', '')
        elif key == 'init_containers':
            containers = []
            for c in capsule.init_containers:
                container = containers_view.format_container(
                    context, None, c)
                containers.append(container)
            yield ('init_containers', containers)
        elif key == 'containers':
            containers = []
            for c in capsule.containers:
                container = containers_view.format_container(
                    context, None, c)
                containers.append(container)
            yield ('containers', containers)
        elif key == 'name':
            if legacy_api_version:
                yield('meta_name', value)
            else:
                yield(key, value)
        elif key == 'labels':
            if legacy_api_version:
                yield('meta_labels', value)
            else:
                yield(key, value)
        elif key == 'restart_policy':
            if legacy_api_version:
                if 'Name' in value:
                    yield(key, value['Name'])
            else:
                yield(key, value)
        else:
            yield (key, value)

    return dict(itertools.chain.from_iterable(
        transform(k, v) for k, v in capsule.as_dict().items()))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Add quota usage

Revision ID: 012a730926e8
Revises: 3298c6a5c3d9
Create Date: 2018-05-30 21:23:17.659203

"""

# revision identifiers, used by Alembic.
revision = '012a730926e8'
down_revision = '3298c6a5c3d9'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'quota_usages',
        sa.Column('created_at', sa.DateTime()),
        sa.Column('updated_at', sa.DateTime()),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('project_id', sa.String(length=255), index=True),
        sa.Column('resource', sa.String(length=255), nullable=False),
        sa.Column('in_use', sa.Integer, nullable=False),
        sa.Column('reserved', sa.Integer, nullable=False),
        sa.Column('until_refresh', sa.Integer),
        mysql_engine='InnoDB'
    )



#    Copyright 2016 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from io import StringIO
from oslo_utils import uuidutils

from zun.common import consts
from zun.common import exception
from zun.compute import claims
from zun.compute import manager
import zun.conf
from zun import objects
from zun.objects.container import Container
from zun.objects.container_action import ContainerAction
from zun.objects.container_action import ContainerActionEvent
from zun.objects.exec_instance import ExecInstance
from zun.objects.image import Image
from zun.objects.volume_mapping import VolumeMapping
from zun.objects.zun_network import ZunNetwork
from zun.tests import base
from zun.tests.unit.container.fake_driver import FakeDriver as fake_driver
from zun.tests.unit.db import utils


class FakeResourceTracker(object):

    def __init__(self, *args, **kwargs):
        self.compute_node = mock.MagicMock()

    def container_claim(self, context, container, pci_requests, limits):
        return claims.NopClaim()

    def container_update_claim(self, context, container, old_container,
                               limits):
        return claims.NopClaim()

    def remove_usage_from_container(self, contxt, context, is_remmoved=True):
        pass


class FakeVolumeMapping(object):

    def __init__(self):
        self.__class__.volumes = []
        self.volume_provider = 'fake_provider'
        self.container_path = 'fake_path'
        self.container_uuid = 'fake-cid'
        self.cinder_volume_id = 'fake-vid'
        self.volume_id = 123
        self.connection_info = None
        self.auto_remove = False
        self.uuid = 'fake-uuid'
        self.volume = mock.Mock()
        self.volume.refresh.side_effect = self._volume_refresh

    def _volume_refresh(self):
        self.connection_info = 'refresh-info'

    def create(self, context):
        self.__class__.volumes.append(self)

    def destroy(self):
        self.__class__.volumes.remove(self)

    def obj_attr_is_set(self, name):
        return bool(self.__class__.volumes)

    @classmethod
    def list_by_container(cls, context, container_id):
        return cls.volumes

    @classmethod
    def list_by_cinder_volume(cls, context, volume_id):
        return cls.volumes

    @classmethod
    def count(cls, context, **filters):
        return len(cls.volumes)


class TestManager(base.TestCase):

    def setUp(self):
        super(TestManager, self).setUp()
        p = mock.patch('zun.scheduler.client.report.SchedulerReportClient')
        p.start()
        self.addCleanup(p.stop)

        zun.conf.CONF.set_override('container_driver', 'fake')
        zun.conf.CONF.set_override('capsule_driver', 'fake')
        self.compute_manager = manager.Manager()
        self.compute_manager._resource_tracker = FakeResourceTracker()

    @mock.patch.object(Container, 'save')
    def test_init_container_sets_creating_error(self, mock_save):
        container = Container(self.context, **utils.get_test_container())
        container.status = consts.CREATING
        self.compute_manager._init_container(context=self.context,
                                             container=container)
        self.assertEqual(consts.ERROR, container.status)
        self.assertIsNone(container.task_state)

    @mock.patch.object(Container, 'save')
    def test_init_container_sets_creating_tasks_error(self, mock_save):
        tasks = [consts.CONTAINER_CREATING, consts.IMAGE_PULLING]
        container = Container(self.context, **utils.get_test_container())
        for task in tasks:
            container.task_state = task
            self.compute_manager._init_container(context=self.context,
                                                 container=container)
            self.assertEqual(consts.ERROR, container.status)
            self.assertIsNone(container.task_state)

    @mock.patch.object(manager.Manager, 'container_reboot')
    @mock.patch.object(Container, 'save')
    def test_init_container_retries_reboot(self, mock_save,
                                           mock_container_reboot):
        container = Container(self.context, **utils.get_test_container())
        container.task_state = consts.CONTAINER_REBOOTING
        self.compute_manager._init_container(self.context, container)
        mock_container_reboot.assert_called_once_with(self.context,
                                                      container, 60)

    @mock.patch.object(manager.Manager, 'container_start')
    @mock.patch.object(Container, 'save')
    def test_init_container_retries_start(self, mock_save,
                                          mock_container_start):
        container = Container(self.context, **utils.get_test_container())
        container.task_state = consts.CONTAINER_STARTING
        container.status = consts.STOPPED
        self.compute_manager._init_container(self.context, container)
        mock_container_start.assert_called_once_with(self.context,
                                                     container)

    @mock.patch.object(manager.Manager, 'container_start')
    @mock.patch.object(Container, 'save')
    def test_container_reboot_after_host_reboot(self, mock_save,
                                                mock_container_start):
        container_1 = Container(self.context, **utils.get_test_container())
        container_1.status = consts.RUNNING
        self.compute_manager.restore_running_container(self.context,
                                                       container_1,
                                                       consts.STOPPED)
        mock_container_start.assert_called_once_with(self.context,
                                                     container_1)

    @mock.patch.object(manager.Manager, 'container_stop')
    @mock.patch.object(Container, 'save')
    def test_init_container_retries_stop(self, mock_save,
                                         mock_container_stop):
        container = Container(self.context, **utils.get_test_container())
        container.task_state = consts.CONTAINER_STOPPING
        self.compute_manager._init_container(self.context, container)
        mock_container_stop.assert_called_once_with(self.context,
                                                    container, 60)

    @mock.patch.object(manager.Manager, 'container_delete')
    @mock.patch.object(Container, 'save')
    def test_init_container_retries_deleting(self, mock_save,
                                             mock_container_delete):
        kw = {'status': consts.DELETING,
              'task_state': None}
        container = Container(self.context, **utils.get_test_container(**kw))
        self.compute_manager._init_container(self.context, container)
        mock_container_delete.assert_called_once_with(self.context, container,
                                                      force=True)

    @mock.patch.object(manager.Manager, 'container_delete')
    @mock.patch.object(Container, 'save')
    def test_init_container_retries_container_delete_task(
            self, mock_save, mock_container_delete):
        container = Container(self.context, **utils.get_test_container())
        container.task_state = consts.CONTAINER_DELETING
        self.compute_manager._init_container(self.context, container)
        mock_container_delete.assert_called_once_with(self.context, container,
                                                      force=True)

    @mock.patch.object(Container, 'save')
    def test_fail_container(self, mock_save):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._fail_container(self.context, container,
                                             "Creation Failed")
        self.assertEqual(consts.ERROR, container.status)
        self.assertEqual("Creation Failed", container.status_reason)
        self.assertIsNone(container.task_state)

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(fake_driver, 'create')
    def test_container_create(self, mock_create, mock_pull, mock_save,
                              mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        image = {'image': 'repo', 'path': 'out_path', 'driver': 'glance'}
        mock_pull.return_value = image, False
        self.compute_manager._resource_tracker = FakeResourceTracker()
        networks = []
        volumes = []
        self.compute_manager._do_container_create(self.context, container,
                                                  networks, volumes)
        mock_save.assert_called_with(self.context)
        mock_pull.assert_any_call(self.context, container.image, '',
                                  'always', 'glance', registry=None)
        mock_create.assert_called_once_with(self.context, container, image,
                                            networks, volumes)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_create'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(manager.Manager, '_fail_container')
    def test_container_create_pull_image_failed_docker_error(
            self, mock_fail, mock_pull, mock_save, mock_event_finish,
            mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_pull.side_effect = exception.DockerError("Pull Failed")
        networks = []
        volumes = []
        self.assertRaises(exception.DockerError,
                          self.compute_manager._do_container_create,
                          self.context, container, networks, volumes)
        mock_fail.assert_called_once_with(self.context,
                                          container, "Pull Failed")
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_create'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(manager.Manager, '_fail_container')
    def test_container_create_pull_image_failed_image_not_found(
            self, mock_fail, mock_pull, mock_save, mock_event_finish,
            mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_pull.side_effect = exception.ImageNotFound("Image Not Found")
        networks = []
        volumes = []
        self.assertRaises(exception.ImageNotFound,
                          self.compute_manager._do_container_create,
                          self.context, container, networks, volumes)
        mock_fail.assert_called_once_with(self.context,
                                          container, "Image Not Found")
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_create'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(manager.Manager, '_fail_container')
    def test_container_create_pull_image_failed_zun_exception(
            self, mock_fail, mock_pull, mock_save, mock_event_finish,
            mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_pull.side_effect = exception.ZunException(
            message="Image Not Found")
        networks = []
        volumes = []
        self.assertRaises(
            exception.ZunException,
            self.compute_manager._do_container_create,
            self.context, container, networks, volumes)
        mock_fail.assert_called_once_with(self.context,
                                          container, "Image Not Found")
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_create'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(fake_driver, 'create')
    @mock.patch.object(manager.Manager, '_fail_container')
    def test_container_create_docker_create_failed(
            self, mock_fail, mock_create, mock_pull, mock_save,
            mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        image = {'image': 'repo', 'path': 'out_path', 'driver': 'glance',
                 'repo': 'test', 'tag': 'testtag'}
        mock_pull.return_value = image, False
        mock_create.side_effect = exception.DockerError("Creation Failed")
        self.compute_manager._resource_tracker = FakeResourceTracker()
        networks = []
        volumes = []
        self.assertRaises(exception.DockerError,
                          self.compute_manager._do_container_create,
                          self.context, container, networks, volumes)
        mock_fail.assert_called_once_with(
            self.context, container, "Creation Failed", unset_host=True)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_create'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch('zun.common.utils.spawn_n')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(VolumeMapping, 'count',
                       side_effect=FakeVolumeMapping.count)
    @mock.patch.object(VolumeMapping, 'list_by_container',
                       side_effect=FakeVolumeMapping.list_by_container)
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(fake_driver, 'detach_volume')
    @mock.patch.object(fake_driver, 'attach_volume')
    @mock.patch.object(fake_driver, 'is_volume_available')
    @mock.patch.object(fake_driver, 'create')
    @mock.patch.object(fake_driver, 'start')
    def test_container_run(
            self, mock_start, mock_create,
            mock_is_volume_available, mock_attach_volume,
            mock_detach_volume, mock_pull, mock_list_by_container, mock_count,
            mock_save, mock_spawn_n, mock_action_finish, mock_event_finish,
            mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        image = {'image': 'repo', 'path': 'out_path', 'driver': 'glance'}
        mock_create.return_value = container
        mock_pull.return_value = image, False
        mock_is_volume_available.return_value = True, False
        mock_spawn_n.side_effect = lambda f, *x, **y: f(*x, **y)
        container.status = 'Stopped'
        self.compute_manager._resource_tracker = FakeResourceTracker()
        networks = []
        volumes = {container.uuid: [FakeVolumeMapping()]}
        self.compute_manager.container_create(
            self.context,
            requested_networks=networks,
            requested_volumes=volumes,
            container=container,
            limits=None, run=True)
        mock_save.assert_called_with(self.context)
        mock_pull.assert_any_call(self.context, container.image, '',
                                  'always', 'glance', registry=None)
        mock_create.assert_called_once_with(self.context, container, image,
                                            networks, volumes)
        mock_start.assert_called_once_with(self.context, container)
        mock_attach_volume.assert_called_once()
        mock_detach_volume.assert_not_called()
        mock_is_volume_available.assert_called_once()
        self.assertEqual(1, len(FakeVolumeMapping.volumes))

    @mock.patch.object(fake_driver, 'delete_volume')
    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch('zun.common.utils.spawn_n')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(VolumeMapping, 'count',
                       side_effect=FakeVolumeMapping.count)
    @mock.patch.object(VolumeMapping, 'list_by_cinder_volume',
                       side_effect=FakeVolumeMapping.list_by_cinder_volume)
    @mock.patch.object(VolumeMapping, 'list_by_container',
                       side_effect=FakeVolumeMapping.list_by_container)
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(fake_driver, 'detach_volume')
    @mock.patch.object(fake_driver, 'attach_volume')
    @mock.patch.object(fake_driver, 'is_volume_available')
    @mock.patch.object(fake_driver, 'create')
    @mock.patch.object(fake_driver, 'start')
    def test_container_run_driver_attach_failed(
            self, mock_start, mock_create,
            mock_is_volume_available, mock_attach_volume,
            mock_detach_volume, mock_pull, mock_list_by_container,
            mock_list_by_volume, mock_count, mock_save,
            mock_spawn_n, mock_action_finish, mock_event_finish,
            mock_event_start, mock_delete_volume):
        mock_is_volume_available.return_value = True, False
        mock_attach_volume.side_effect = [None, base.TestingException("fake")]
        container = Container(self.context, **utils.get_test_container())
        vol = FakeVolumeMapping()
        vol.auto_remove = True
        vol2 = FakeVolumeMapping()
        vol2.auto_remove = True
        image = {'image': 'repo', 'path': 'out_path', 'driver': 'glance'}
        mock_create.return_value = container
        mock_pull.return_value = image, False
        mock_spawn_n.side_effect = lambda f, *x, **y: f(*x, **y)
        container.status = 'Stopped'
        self.compute_manager._resource_tracker = FakeResourceTracker()
        networks = []
        volumes = {container.uuid: [vol, vol2]}
        self.assertRaises(
            base.TestingException,
            self.compute_manager.container_create,
            self.context,
            requested_networks=networks,
            requested_volumes=volumes,
            container=container,
            limits=None, run=True)
        mock_save.assert_called_with(self.context)
        mock_pull.assert_not_called()
        mock_create.assert_not_called()
        mock_start.assert_not_called()
        mock_attach_volume.assert_has_calls([
            mock.call(mock.ANY, vol), mock.call(mock.ANY, vol2)])
        mock_detach_volume.assert_has_calls([
            mock.call(mock.ANY, vol)])
        self.assertEqual(0, len(FakeVolumeMapping.volumes))

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch('zun.common.utils.spawn_n')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(VolumeMapping, 'count',
                       side_effect=FakeVolumeMapping.count)
    @mock.patch.object(VolumeMapping, 'list_by_cinder_volume',
                       side_effect=FakeVolumeMapping.list_by_cinder_volume)
    @mock.patch.object(VolumeMapping, 'list_by_container',
                       side_effect=FakeVolumeMapping.list_by_container)
    @mock.patch.object(fake_driver, 'detach_volume')
    @mock.patch.object(fake_driver, 'attach_volume')
    @mock.patch.object(fake_driver, 'is_volume_available')
    @mock.patch.object(fake_driver, 'pull_image')
    def test_container_run_image_not_found(
            self, mock_pull, mock_is_volume_available,
            mock_attach_volume, mock_detach_volume,
            mock_list_by_container, mock_list_by_volume, mock_count,
            mock_save, mock_spawn_n, mock_action_finish, mock_event_finish,
            mock_event_start):
        container_dict = utils.get_test_container(
            image='test:latest', image_driver='docker',
            image_pull_policy='ifnotpresent')
        container = Container(self.context, **container_dict)
        mock_is_volume_available.return_value = True, False
        mock_pull.side_effect = exception.ImageNotFound(
            message="Image Not Found")
        mock_spawn_n.side_effect = lambda f, *x, **y: f(*x, **y)
        networks = []
        volumes = {container.uuid: [FakeVolumeMapping()]}
        self.assertRaises(
            exception.ImageNotFound,
            self.compute_manager.container_create,
            self.context,
            requested_networks=networks,
            requested_volumes=volumes,
            container=container,
            limits=None, run=True)
        mock_save.assert_called_with(self.context)
        self.assertEqual('Error', container.status)
        self.assertEqual('Image Not Found', container.status_reason)
        mock_pull.assert_called_once_with(self.context, 'test', 'latest',
                                          'ifnotpresent', 'docker',
                                          registry=None)
        mock_attach_volume.assert_called_once()
        mock_detach_volume.assert_called_once()
        mock_is_volume_available.assert_called_once()
        self.assertEqual(0, len(FakeVolumeMapping.volumes))

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch('zun.common.utils.spawn_n')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(VolumeMapping, 'count',
                       side_effect=FakeVolumeMapping.count)
    @mock.patch.object(VolumeMapping, 'list_by_cinder_volume',
                       side_effect=FakeVolumeMapping.list_by_cinder_volume)
    @mock.patch.object(VolumeMapping, 'list_by_container',
                       side_effect=FakeVolumeMapping.list_by_container)
    @mock.patch.object(fake_driver, 'detach_volume')
    @mock.patch.object(fake_driver, 'attach_volume')
    @mock.patch.object(fake_driver, 'is_volume_available')
    @mock.patch.object(fake_driver, 'pull_image')
    def test_container_run_image_pull_exception_raised(
            self, mock_pull, mock_is_volume_available,
            mock_attach_volume, mock_detach_volume,
            mock_list_by_container, mock_list_by_volume, mock_count,
            mock_save, mock_spawn_n, mock_action_finish, mock_event_finish,
            mock_event_start):
        container_dict = utils.get_test_container(
            image='test:latest', image_driver='docker',
            image_pull_policy='ifnotpresent')
        container = Container(self.context, **container_dict)
        mock_is_volume_available.return_value = True, False
        mock_pull.side_effect = exception.ZunException(
            message="Image Not Found")
        mock_spawn_n.side_effect = lambda f, *x, **y: f(*x, **y)
        networks = []
        volumes = {container.uuid: [FakeVolumeMapping()]}
        self.assertRaises(
            exception.ZunException,
            self.compute_manager.container_create,
            self.context,
            requested_networks=networks,
            requested_volumes=volumes,
            container=container,
            limits=None, run=True)
        mock_save.assert_called_with(self.context)
        self.assertEqual('Error', container.status)
        self.assertEqual('Image Not Found', container.status_reason)
        mock_pull.assert_called_once_with(self.context, 'test', 'latest',
                                          'ifnotpresent', 'docker',
                                          registry=None)
        mock_attach_volume.assert_called_once()
        mock_detach_volume.assert_called_once()
        mock_is_volume_available.assert_called_once()
        self.assertEqual(0, len(FakeVolumeMapping.volumes))

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch('zun.common.utils.spawn_n')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(VolumeMapping, 'count',
                       side_effect=FakeVolumeMapping.count)
    @mock.patch.object(VolumeMapping, 'list_by_cinder_volume',
                       side_effect=FakeVolumeMapping.list_by_cinder_volume)
    @mock.patch.object(VolumeMapping, 'list_by_container',
                       side_effect=FakeVolumeMapping.list_by_container)
    @mock.patch.object(fake_driver, 'detach_volume')
    @mock.patch.object(fake_driver, 'attach_volume')
    @mock.patch.object(fake_driver, 'is_volume_available')
    @mock.patch.object(fake_driver, 'pull_image')
    def test_container_run_image_pull_docker_error(
            self, mock_pull, mock_is_volume_available,
            mock_attach_volume, mock_detach_volume,
            mock_list_by_container, mock_list_by_volume, mock_count,
            mock_save, mock_spawn_n, mock_action_finish, mock_event_finish,
            mock_event_start):
        container_dict = utils.get_test_container(
            image='test:latest', image_driver='docker',
            image_pull_policy='ifnotpresent')
        container = Container(self.context, **container_dict)
        mock_is_volume_available.return_value = True, False
        mock_pull.side_effect = exception.DockerError(
            message="Docker Error occurred")
        mock_spawn_n.side_effect = lambda f, *x, **y: f(*x, **y)
        networks = []
        volumes = {container.uuid: [FakeVolumeMapping()]}
        self.assertRaises(
            exception.DockerError,
            self.compute_manager.container_create,
            self.context,
            requested_networks=networks,
            requested_volumes=volumes,
            container=container,
            limits=None, run=True)
        mock_save.assert_called_with(self.context)
        self.assertEqual('Error', container.status)
        self.assertEqual('Docker Error occurred', container.status_reason)
        mock_pull.assert_called_once_with(self.context, 'test', 'latest',
                                          'ifnotpresent', 'docker',
                                          registry=None)
        mock_attach_volume.assert_called_once()
        mock_detach_volume.assert_called_once()
        mock_is_volume_available.assert_called_once()
        self.assertEqual(0, len(FakeVolumeMapping.volumes))

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch('zun.common.utils.spawn_n')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(VolumeMapping, 'count',
                       side_effect=FakeVolumeMapping.count)
    @mock.patch.object(VolumeMapping, 'list_by_cinder_volume',
                       side_effect=FakeVolumeMapping.list_by_cinder_volume)
    @mock.patch.object(VolumeMapping, 'list_by_container',
                       side_effect=FakeVolumeMapping.list_by_container)
    @mock.patch.object(fake_driver, 'detach_volume')
    @mock.patch.object(fake_driver, 'attach_volume')
    @mock.patch.object(fake_driver, 'is_volume_available')
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(fake_driver, 'create')
    def test_container_run_create_raises_docker_error(
            self, mock_create, mock_pull, mock_is_volume_available,
            mock_attach_volume, mock_detach_volume,
            mock_list_by_container, mock_list_by_volume, mock_count,
            mock_save, mock_spawn_n, mock_action_finish,
            mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        image = {'image': 'repo', 'path': 'out_path', 'driver': 'glance',
                 'repo': 'test', 'tag': 'testtag'}
        mock_pull.return_value = image, True
        mock_is_volume_available.return_value = True, False
        mock_create.side_effect = exception.DockerError(
            message="Docker Error occurred")
        mock_spawn_n.side_effect = lambda f, *x, **y: f(*x, **y)
        self.compute_manager._resource_tracker = FakeResourceTracker()
        networks = []
        volumes = {container.uuid: [FakeVolumeMapping()]}
        self.assertRaises(
            exception.DockerError,
            self.compute_manager.container_create,
            self.context,
            requested_networks=networks,
            requested_volumes=volumes,
            container=container,
            limits=None, run=True)
        mock_save.assert_called_with(self.context)
        self.assertEqual('Error', container.status)
        self.assertEqual('Docker Error occurred', container.status_reason)
        mock_pull.assert_any_call(self.context, container.image, '',
                                  'always', 'glance', registry=None)
        mock_create.assert_called_once_with(
            self.context, container, image, networks, volumes)
        mock_attach_volume.assert_called_once()
        mock_detach_volume.assert_called_once()
        mock_is_volume_available.assert_called_once()
        self.assertEqual(0, len(FakeVolumeMapping.volumes))

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch('zun.common.utils.spawn_n')
    @mock.patch.object(objects.Capsule, 'save')
    @mock.patch.object(objects.CapsuleContainer, 'list_by_capsule_id')
    @mock.patch.object(objects.CapsuleInitContainer, 'list_by_capsule_id')
    @mock.patch.object(VolumeMapping, 'count',
                       side_effect=FakeVolumeMapping.count)
    @mock.patch.object(VolumeMapping, 'list_by_container',
                       side_effect=FakeVolumeMapping.list_by_container)
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(fake_driver, 'detach_volume')
    @mock.patch.object(fake_driver, 'attach_volume')
    @mock.patch.object(fake_driver, 'is_volume_available')
    @mock.patch.object(fake_driver, 'create_capsule')
    @mock.patch.object(fake_driver, 'start')
    def test_capsule_create(
            self, mock_start, mock_create,
            mock_is_volume_available, mock_attach_volume,
            mock_detach_volume, mock_pull, mock_list_by_container, mock_count,
            mock_init_container_list, mock_capsule_container_list,
            mock_save, mock_spawn_n, mock_action_finish, mock_event_finish,
            mock_event_start):
        capsule = objects.Capsule(self.context, **utils.get_test_container())
        image = {'image': 'repo', 'path': 'out_path', 'driver': 'glance'}
        mock_create.return_value = capsule
        container_uuid = uuidutils.generate_uuid()
        container = objects.CapsuleContainer(
            self.context, **utils.get_test_container(uuid=container_uuid))
        mock_capsule_container_list.return_value = [container]
        container_uuid = uuidutils.generate_uuid()
        init_container = objects.CapsuleInitContainer(
            self.context, **utils.get_test_container(uuid=container_uuid))
        mock_init_container_list.return_value = [init_container]
        mock_pull.return_value = image, False
        mock_is_volume_available.return_value = True, False
        mock_spawn_n.side_effect = lambda f, *x, **y: f(*x, **y)
        capsule.status = 'Running'
        self.compute_manager._resource_tracker = FakeResourceTracker()
        networks = []
        volmap1 = FakeVolumeMapping()
        volmap1.uuid = 'fake-uuid-1'
        volmap2 = FakeVolumeMapping()
        volmap2.uuid = 'fake-uuid-2'
        volumes = {container.uuid: [volmap1],
                   init_container.uuid: [volmap2]}

        def attach_volume(context, volmap):
            volmap.connection_info = 'fake-info'

        mock_attach_volume.side_effect = attach_volume
        self.compute_manager.container_create(
            self.context,
            requested_networks=networks,
            requested_volumes=volumes,
            container=capsule,
            limits=None, run=True)
        mock_save.assert_called_with(self.context)
        mock_pull.assert_any_call(self.context, capsule.image, '',
                                  'always', 'glance', registry=None)
        mock_create.assert_called_once_with(self.context, capsule, image,
                                            networks, volumes)
        mock_start.assert_called_once_with(self.context, capsule)
        mock_attach_volume.assert_called_once()
        mock_detach_volume.assert_not_called()
        self.assertEqual(2, mock_is_volume_available.call_count)
        self.assertEqual(2, len(FakeVolumeMapping.volumes))

    @mock.patch.object(FakeResourceTracker,
                       'remove_usage_from_container')
    @mock.patch.object(Container, 'destroy')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(VolumeMapping, 'list_by_container')
    @mock.patch.object(fake_driver, 'delete')
    def test_container_delete(
            self, mock_delete, mock_list_by_container, mock_save,
            mock_cnt_destroy, mock_remove_usage):
        mock_list_by_container.return_value = []
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._do_container_delete(self. context, container,
                                                  False)
        mock_save.assert_called_with(self.context)
        mock_delete.assert_called_once_with(self.context, container, False)
        mock_cnt_destroy.assert_called_once_with(self.context)
        mock_remove_usage.assert_called_once_with(self.context, container,
                                                  True)

    @mock.patch.object(FakeResourceTracker,
                       'remove_usage_from_container')
    @mock.patch.object(Container, 'destroy')
    @mock.patch.object(manager.Manager, '_fail_container')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'delete')
    def test_container_delete_failed(self, mock_delete, mock_save,
                                     mock_fail, mock_destroy,
                                     mock_remove_usage):
        container = Container(self.context, **utils.get_test_container())
        mock_delete.side_effect = exception.DockerError(
            message="Docker Error occurred")
        self.assertRaises(exception.DockerError,
                          self.compute_manager._do_container_delete,
                          self.context, container, False)
        mock_save.assert_called_with(self.context)
        mock_fail.assert_called_with(self.context,
                                     container, 'Docker Error occurred')
        mock_destroy.assert_not_called()
        mock_remove_usage.assert_not_called()

    @mock.patch.object(FakeResourceTracker,
                       'remove_usage_from_container')
    @mock.patch.object(Container, 'destroy')
    @mock.patch.object(manager.Manager, '_fail_container')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(VolumeMapping, 'list_by_container')
    @mock.patch.object(fake_driver, 'delete')
    def test_container_delete_failed_force(self, mock_delete,
                                           mock_list_by_container,
                                           mock_save,
                                           mock_fail, mock_destroy,
                                           mock_remove_usage):
        mock_list_by_container.return_value = []
        container = Container(self.context, **utils.get_test_container())
        mock_delete.side_effect = exception.DockerError(
            message="Docker Error occurred")
        self.compute_manager._do_container_delete(self.context, container,
                                                  True)
        mock_save.assert_called_with(self.context)
        mock_fail.assert_called_with(self.context,
                                     container, 'Docker Error occurred')
        mock_destroy.assert_called_once_with(self.context)
        mock_remove_usage.assert_called_once_with(self.context, container,
                                                  True)

    @mock.patch.object(fake_driver, 'show')
    def test_container_show(self, mock_show):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager.container_show(self.context, container)
        mock_show.assert_called_once_with(self.context, container)

    @mock.patch.object(fake_driver, 'show')
    def test_container_show_failed(self, mock_show):
        container = Container(self.context, **utils.get_test_container())
        mock_show.side_effect = exception.DockerError
        self.assertRaises(exception.DockerError,
                          self.compute_manager.container_show,
                          self.context, container)

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch('zun.compute.manager.Manager._get_vol_info')
    @mock.patch('zun.compute.manager.Manager._get_network_info')
    @mock.patch.object(fake_driver, 'pull_image')
    @mock.patch.object(fake_driver, 'check_container_exist')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'create')
    @mock.patch.object(fake_driver, 'delete')
    def test_container_rebuild(self, mock_delete, mock_create,
                               mock_save, mock_check, mock_pull,
                               mock_get_network_info, mock_get_vol_info,
                               mock_action_finish,
                               mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        image = {'image': 'repo', 'path': 'out_path', 'driver': 'glance'}
        mock_pull.return_value = image, False
        container.status = 'Created'
        mock_get_network_info.return_value = []
        mock_get_vol_info.return_value = []
        mock_check.return_value = True
        self.compute_manager._do_container_rebuild(
            self.context, container, False)
        mock_save.assert_called_with(self.context)
        self.assertTrue(mock_create.called)
        mock_delete.assert_called_once_with(self.context, container, True)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        mock_action_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_rebuild'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch('zun.compute.manager.Manager._get_vol_info')
    @mock.patch('zun.compute.manager.Manager._get_network_info')
    @mock.patch.object(manager.Manager, '_fail_container')
    def test_container_rebuild_failed(
            self, mock_fail, mock_get_network_info, mock_get_vol_info,
            mock_save, mock_action_finish, mock_event_finish,
            mock_event_start):
        mock_get_vol_info.return_value = []
        fake_exc = exception.PortNotFound(port='fake-port')
        mock_get_network_info.side_effect = fake_exc
        container = Container(self.context, **utils.get_test_container())
        self.assertRaises(exception.PortNotFound,
                          self.compute_manager._do_container_rebuild,
                          self.context, container, True)
        mock_fail.assert_called_with(self.context,
                                     container, str(fake_exc))
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        mock_action_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_rebuild'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'reboot')
    def test_container_reboot(self, mock_reboot, mock_save, mock_event_finish,
                              mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._do_container_reboot(self.context, container, 10)
        mock_save.assert_called_with(self.context)
        mock_reboot.assert_called_once_with(self.context, container, 10)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_reboot'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'reboot')
    def test_container_reboot_failed(self, mock_reboot, mock_save,
                                     mock_event_finish,
                                     mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_reboot.side_effect = exception.DockerError(
            message="Docker Error occurred")
        self.compute_manager._do_container_reboot(self.context, container, 10)
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_reboot'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'stop')
    def test_container_stop(self, mock_stop, mock_save, mock_event_finish,
                            mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._do_container_stop(self.context, container, 10)
        mock_save.assert_called_with(self.context)
        mock_stop.assert_called_once_with(self.context, container, 10)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_stop'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'stop')
    def test_container_stop_failed(self, mock_stop, mock_save,
                                   mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_stop.side_effect = exception.DockerError(
            message="Docker Error occurred")
        self.compute_manager._do_container_stop(self.context, container, 10)
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_stop'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'start')
    def test_container_start(self, mock_start, mock_save, mock_event_finish,
                             mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._do_container_start(self.context, container)
        mock_save.assert_called_with(self.context)
        mock_start.assert_called_once_with(self.context, container)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_start'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(manager.Manager, '_fail_container')
    @mock.patch.object(fake_driver, 'start')
    def test_container_start_failed(self, mock_start,
                                    mock_fail, mock_save, mock_event_finish,
                                    mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_start.side_effect = exception.DockerError(
            message="Docker Error occurred")
        self.assertRaises(exception.DockerError,
                          self.compute_manager._do_container_start,
                          self.context, container)
        mock_save.assert_called_with(self.context)
        mock_fail.assert_called_with(self.context,
                                     container, 'Docker Error occurred')
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_start'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'pause')
    def test_container_pause(self, mock_pause, mock_save,
                             mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._do_container_pause(self.context, container)
        mock_pause.assert_called_once_with(self.context, container)
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_pause'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'pause')
    def test_container_pause_failed(self, mock_pause, mock_save,
                                    mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_pause.side_effect = exception.DockerError(
            message="Docker Error occurred")
        self.compute_manager._do_container_pause(self.context, container)
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_pause'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'unpause')
    def test_container_unpause(self, mock_unpause, mock_save,
                               mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._do_container_unpause(self.context, container)
        mock_unpause.assert_called_once_with(self.context, container)
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_unpause'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'unpause')
    def test_container_unpause_failed(self, mock_unpause, mock_save,
                                      mock_event_finish,
                                      mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_unpause.side_effect = exception.DockerError(
            message="Docker Error occurred")
        self.compute_manager._do_container_unpause(self.context, container)
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_unpause'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(fake_driver, 'show_logs')
    def test_container_logs(self, mock_logs):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager.container_logs(self.context,
                                            container, True, True,
                                            False, 'all', None)
        mock_logs.assert_called_once_with(
            self.context, container, stderr=True, stdout=True,
            timestamps=False, tail='all', since=None)

    @mock.patch.object(fake_driver, 'show_logs')
    def test_container_logs_failed(self, mock_logs):
        container = Container(self.context, **utils.get_test_container())
        mock_logs.side_effect = exception.DockerError
        self.assertRaises(exception.DockerError,
                          self.compute_manager.container_logs,
                          self.context, container, True, True,
                          False, 'all', None)

    @mock.patch.object(fake_driver, 'execute_run')
    @mock.patch.object(fake_driver, 'execute_create')
    def test_container_execute(self, mock_execute_create, mock_execute_run):
        mock_execute_create.return_value = 'fake_exec_id'
        mock_execute_run.return_value = 'fake_output', 'fake_exit_code'
        container = Container(self.context, **utils.get_test_container())
        result = self.compute_manager.container_exec(
            self.context, container, 'fake_cmd', True, False)
        self.assertEqual('fake_output', result.get('output'))
        self.assertEqual('fake_exit_code', result.get('exit_code'))
        self.assertIsNone(result.get('exec_id'))
        self.assertIsNone(result.get('token'))
        mock_execute_create.assert_called_once_with(
            self.context, container, 'fake_cmd', False)
        mock_execute_run.assert_called_once_with('fake_exec_id', 'fake_cmd')

    @mock.patch.object(ExecInstance, 'create')
    @mock.patch.object(fake_driver, 'execute_run')
    @mock.patch.object(fake_driver, 'execute_create')
    def test_container_execute_interactive(
            self, mock_execute_create, mock_execute_run, mock_create):
        mock_execute_create.return_value = 'fake_exec_id'
        container = Container(self.context, **utils.get_test_container())
        result = self.compute_manager.container_exec(
            self.context, container, 'fake_cmd', False, True)
        self.assertIsNone(result.get('output'))
        self.assertIsNone(result.get('exit_code'))
        self.assertEqual('fake_exec_id', result.get('exec_id'))
        self.assertIsNotNone(result.get('token'))
        mock_execute_create.assert_called_once_with(
            self.context, container, 'fake_cmd', True)
        mock_execute_run.assert_not_called()

    @mock.patch.object(fake_driver, 'execute_create')
    def test_container_execute_failed(self, mock_execute_create):
        container = Container(self.context, **utils.get_test_container())
        mock_execute_create.side_effect = exception.DockerError
        self.assertRaises(exception.DockerError,
                          self.compute_manager.container_exec,
                          self.context, container, 'fake_cmd', True, False)

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'kill')
    def test_container_kill(self, mock_kill, mock_save,
                            mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._do_container_kill(self.context, container, None)
        mock_kill.assert_called_once_with(self.context, container, None)
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_kill'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'kill')
    def test_container_kill_failed(self, mock_kill, mock_save,
                                   mock_event_finish,
                                   mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_kill.side_effect = exception.DockerError(
            message="Docker Error occurred")
        self.compute_manager._do_container_kill(self.context, container, None)
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_kill'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'update')
    def test_container_update(self, mock_update, mock_save):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager.container_update(self.context, container,
                                              {'memory': 512})
        mock_save.assert_called_with(self.context)
        mock_update.assert_called_once_with(self.context, container)

    @mock.patch.object(fake_driver, 'update')
    def test_container_update_failed(self, mock_update):
        container = Container(self.context, **utils.get_test_container())
        mock_update.side_effect = exception.DockerError
        self.assertRaises(exception.DockerError,
                          self.compute_manager.container_update,
                          self.context, container, {})

    @mock.patch.object(fake_driver, 'update')
    def test_container_update_failed_resources(self, mock_update):
        container = Container(self.context, **utils.get_test_container())
        mock_update.side_effect = exception.ResourcesUnavailable
        self.assertRaises(exception.ResourcesUnavailable,
                          self.compute_manager.container_update,
                          self.context, container, {})

    @mock.patch.object(fake_driver, 'get_websocket_url')
    @mock.patch.object(Container, 'save')
    def test_container_attach_successful(self, mock_save,
                                         mock_get_websocket_url):
        container = Container(self.context, **utils.get_test_container())
        mock_get_websocket_url.return_value = "ws://test"
        self.compute_manager.container_attach(self.context, container)
        mock_get_websocket_url.assert_called_once_with(self.context, container)
        mock_save.assert_called_once_with(self.context)

    @mock.patch.object(fake_driver, 'get_websocket_url')
    def test_container_attach_failed(self, mock_get_websocket_url):
        container = Container(self.context, **utils.get_test_container())
        mock_get_websocket_url.side_effect = Exception
        self.assertRaises(exception.ZunException,
                          self.compute_manager.container_attach,
                          self.context, container)

    @mock.patch.object(fake_driver, 'resize')
    def test_container_resize(self, mock_resize):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager.container_resize(
            self.context, container, "100", "100")
        mock_resize.assert_called_once_with(
            self.context, container, "100", "100")

    @mock.patch.object(fake_driver, 'resize')
    def test_container_resize_failed(self, mock_resize):
        container = Container(self.context, **utils.get_test_container())
        mock_resize.side_effect = exception.DockerError
        self.assertRaises(exception.DockerError,
                          self.compute_manager.container_resize,
                          self.context, container, "100", "100")

    @mock.patch.object(fake_driver, 'inspect_image')
    @mock.patch.object(Image, 'save')
    @mock.patch.object(fake_driver, 'pull_image')
    def test_image_pull(self, mock_pull, mock_save, mock_inspect):
        image = Image(self.context, **utils.get_test_image())
        ret = {'image': 'repo', 'path': 'out_path', 'driver': 'docker'}
        mock_pull.return_value = ret, True
        mock_inspect.return_value = {'Id': 'fake-id', 'Size': 512,
                                     'RepoTags': ['image1:latest']}
        self.compute_manager._do_image_pull(self.context, image)
        mock_pull.assert_any_call(self.context, image.repo, image.tag,
                                  driver_name='docker')
        mock_save.assert_called_once()
        mock_inspect.assert_called_once_with(image.repo + ":" + image.tag)

    @mock.patch.object(fake_driver, 'load_image')
    @mock.patch.object(fake_driver, 'inspect_image')
    @mock.patch.object(Image, 'save')
    @mock.patch.object(fake_driver, 'pull_image')
    def test_image_pull_not_loaded(self, mock_pull, mock_save,
                                   mock_inspect, mock_load):
        image = Image(self.context, **utils.get_test_image())
        repo_tag = image.repo + ":" + image.tag
        ret = {'image': 'repo', 'path': 'out_path', 'driver': 'docker',
               'tags': ['latest']}
        mock_pull.return_value = ret, False
        mock_inspect.return_value = {'Id': 'fake-id', 'Size': 512,
                                     'RepoTags': ['image1:latest']}
        self.compute_manager._do_image_pull(self.context, image)
        mock_pull.assert_any_call(self.context, image.repo, image.tag,
                                  driver_name='docker')
        mock_save.assert_called_once()
        mock_inspect.assert_called_once_with(repo_tag)
        mock_load.assert_called_once_with(ret['path'])

    @mock.patch.object(fake_driver, 'inspect_image')
    @mock.patch.object(Image, 'save')
    @mock.patch.object(fake_driver, 'pull_image')
    def test_image_pull_tag_is_none(self, mock_pull, mock_save, mock_inspect):
        image = Image(self.context, **utils.get_test_image(tag=None))
        ret = {'image': 'repo', 'path': 'out_path', 'driver': 'docker'}
        mock_pull.return_value = ret, True
        mock_inspect.return_value = {'Id': 'fake-id', 'Size': 512,
                                     'RepoTags': ['image1:latest']}
        self.compute_manager._do_image_pull(self.context, image)
        mock_pull.assert_any_call(self.context, image.repo, None,
                                  driver_name='docker')
        mock_save.assert_called_once()
        mock_inspect.assert_called_once_with(image.repo)

    @mock.patch.object(fake_driver, 'execute_resize')
    def test_container_exec_resize(self, mock_resize):
        self.compute_manager.container_exec_resize(
            self.context, 'fake_exec_id', "100", "100")
        mock_resize.assert_called_once_with('fake_exec_id', "100", "100")

    @mock.patch.object(fake_driver, 'execute_resize')
    def test_container_exec_resize_failed(self, mock_resize):
        mock_resize.side_effect = exception.DockerError
        self.assertRaises(exception.DockerError,
                          self.compute_manager.container_exec_resize,
                          self.context, 'fake_exec_id', "100", "100")

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch.object(fake_driver, 'upload_image_data')
    @mock.patch.object(fake_driver, 'get_image')
    @mock.patch.object(fake_driver, 'commit')
    @mock.patch.object(fake_driver, 'pause')
    @mock.patch.object(fake_driver, 'unpause')
    @mock.patch.object(Container, 'save')
    def test_container_commit(
            self, mock_save, mock_unpause, mock_pause, mock_commit,
            mock_get_image, mock_upload_image_data, mock_action_finish,
            mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container(
            status=consts.PAUSED))
        mock_get_image_response = mock.MagicMock()
        mock_get_image_response.data = StringIO().read()
        mock_get_image.return_value = mock_get_image_response
        mock_upload_image_data.return_value = mock.MagicMock()

        self.compute_manager._do_container_commit(self.context,
                                                  mock_get_image_response,
                                                  container, 'repo', 'tag')
        mock_commit.assert_called_once_with(
            self.context, container, 'repo', 'tag')
        mock_pause.assert_not_called()
        mock_unpause.assert_not_called()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_commit'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch.object(fake_driver, 'upload_image_data')
    @mock.patch.object(fake_driver, 'get_image')
    @mock.patch.object(fake_driver, 'commit')
    @mock.patch.object(fake_driver, 'pause')
    @mock.patch.object(fake_driver, 'unpause')
    @mock.patch.object(Container, 'save')
    def test_container_commit_with_pause(
            self, mock_save, mock_unpause, mock_pause, mock_commit,
            mock_get_image, mock_upload_image_data, mock_action_finish,
            mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_get_image_response = mock.MagicMock()
        mock_get_image_response.data = StringIO().read()
        mock_get_image.return_value = mock_get_image_response
        mock_upload_image_data.return_value = mock.MagicMock()
        mock_unpause.return_value = container
        mock_pause.return_value = container

        self.compute_manager._do_container_commit(self.context,
                                                  mock_get_image_response,
                                                  container, 'repo', 'tag')
        mock_commit.assert_called_once_with(
            self.context, container, 'repo', 'tag')
        mock_pause.assert_called_once_with(self.context, container)
        mock_unpause.assert_called_once_with(self.context, container)
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_commit'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(ContainerAction, 'action_finish')
    @mock.patch.object(fake_driver, 'delete_committed_image')
    @mock.patch.object(fake_driver, 'commit')
    @mock.patch.object(fake_driver, 'pause')
    @mock.patch.object(fake_driver, 'unpause')
    @mock.patch.object(Container, 'save')
    def test_container_commit_failed(self, mock_save, mock_unpause, mock_pause,
                                     mock_commit, mock_delete,
                                     mock_action_finish,
                                     mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        mock_get_image_response = mock.MagicMock()
        mock_get_image_response.data = StringIO().read()
        mock_commit.side_effect = exception.DockerError
        mock_unpause.return_value = container
        mock_pause.return_value = container
        self.assertRaises(exception.DockerError,
                          self.compute_manager._do_container_commit,
                          self.context, mock_get_image_response, container,
                          'repo', 'tag')
        self.assertTrue(mock_delete.called)
        mock_commit.assert_called_once_with(
            self.context, container, 'repo', 'tag')
        mock_pause.assert_called_once_with(self.context, container)
        mock_unpause.assert_called_once_with(self.context, container)
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_container_commit'),
            mock_event_finish.call_args[0])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNotNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'network_detach')
    def test_container_network_detach(self, mock_detach, mock_save,
                                      mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._do_network_detach(self.context, container,
                                                'network')
        mock_detach.assert_called_once_with(self.context, container, mock.ANY)
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_network_detach'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(ContainerActionEvent, 'event_start')
    @mock.patch.object(ContainerActionEvent, 'event_finish')
    @mock.patch.object(Container, 'save')
    @mock.patch.object(fake_driver, 'network_attach')
    def test_container_network_attach(self, mock_attach, mock_save,
                                      mock_event_finish, mock_event_start):
        container = Container(self.context, **utils.get_test_container())
        self.compute_manager._do_network_attach(self.context, container,
                                                'network')
        mock_save.assert_called_with(self.context)
        mock_event_start.assert_called_once()
        mock_event_finish.assert_called_once()
        self.assertEqual(
            (self.context, container.uuid, 'compute__do_network_attach'),
            mock_event_finish.call_args[0])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_val'])
        self.assertIsNone(mock_event_finish.call_args[1]['exc_tb'])

    @mock.patch.object(fake_driver, 'is_volume_available')
    @mock.patch.object(manager.Manager, '_fail_container')
    def test_wait_for_volumes_available(self, mock_fail,
                                        mock_is_volume_available):
        mock_is_volume_available.return_value = True, False
        container = Container(self.context, **utils.get_test_container())
        FakeVolumeMapping.container_uuid = container.uuid
        volumes = {container.uuid: [FakeVolumeMapping()]}
        self.compute_manager._wait_for_volumes_available(self.context,
                                                         volumes,
                                                         container)
        mock_is_volume_available.assert_called_once()
        mock_fail.assert_not_called()

    @mock.patch.object(fake_driver, 'delete_volume')
    @mock.patch.object(fake_driver, 'is_volume_available')
    @mock.patch.object(manager.Manager, '_fail_container')
    def test_wait_for_volumes_available_failed(self, mock_fail,
                                               mock_is_volume_available,
                                               mock_delete_volume):
        mock_is_volume_available.return_value = False, True
        container = Container(self.context, **utils.get_test_container())
        FakeVolumeMapping.container_uuid = container.uuid
        volume = FakeVolumeMapping()
        volume.auto_remove = True
        volumes = {container.uuid: [volume]}
        self.assertRaises(exception.Conflict,
                          self.compute_manager._wait_for_volumes_available,
                          self.context, volumes, container, timeout=2)
        self.assertTrue(mock_is_volume_available.called)
        self.assertTrue(mock_fail.called)
        self.assertTrue(mock_delete_volume.called)

    @mock.patch.object(fake_driver, 'create_network')
    def test_network_create(self, mock_create):
        network = ZunNetwork(self.context, **utils.get_test_network())
        ret = ({'Id': '0eeftestnetwork'})
        mock_create.return_value = ret
        self.compute_manager.network_create(self.context, network)
        mock_create.assert_any_call(self.context, network)

    @mock.patch.object(fake_driver, 'delete_network')
    def test_network_delete(self, mock_delete):
        network = ZunNetwork(self.context, **utils.get_test_network())
        self.compute_manager.network_delete(self.context, network)
        mock_delete.assert_any_call(self.context, network)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from testtools.matchers import HasLength

from zun import objects
from zun.tests.unit.db import base
from zun.tests.unit.db import utils


class TestZunServiceObject(base.DbTestCase):

    def setUp(self):
        super(TestZunServiceObject, self).setUp()
        self.fake_zun_service = utils.get_test_zun_service()

    def test_get_by_host_and_binary(self):
        with mock.patch.object(self.dbapi,
                               'get_zun_service',
                               autospec=True) as mock_get_zun_service:
            mock_get_zun_service.return_value = self.fake_zun_service
            ms = objects.ZunService.get_by_host_and_binary(self.context,
                                                           'fake-host',
                                                           'fake-bin')
            mock_get_zun_service.assert_called_once_with('fake-host',
                                                         'fake-bin')
            self.assertEqual(self.context, ms._context)

    def test_get_by_host_and_binary_no_service(self):
        with mock.patch.object(self.dbapi, 'create_zun_service',
                               autospec=True) as mock_get_zun_service:
            mock_get_zun_service.return_value = None
            ms = objects.ZunService.get_by_host_and_binary(self.context,
                                                           'fake-host',
                                                           'fake-bin')

            self.assertIsNone(ms)

    def test_list(self):
        with mock.patch.object(self.dbapi, 'list_zun_services',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_zun_service]
            services = objects.ZunService.list(self.context)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(services, HasLength(1))
            self.assertIsInstance(services[0], objects.ZunService)
            self.assertEqual(self.context, services[0]._context)

    def test_list_by_binary(self):
        with mock.patch.object(self.dbapi, 'list_zun_services_by_binary',
                               autospec=True) as mock_service_list:
            mock_service_list.return_value = [self.fake_zun_service]
            services = objects.ZunService.list_by_binary(self.context, 'bin')
            self.assertEqual(1, mock_service_list.call_count)
            self.assertThat(services, HasLength(1))
            self.assertIsInstance(services[0], objects.ZunService)
            self.assertEqual(self.context, services[0]._context)

    def test_create(self):
        with mock.patch.object(self.dbapi, 'create_zun_service',
                               autospec=True) as mock_create_zun_service:
            mock_create_zun_service.return_value = self.fake_zun_service
            ms_dict = {'host': 'fakehost', 'binary': 'fake-bin'}
            ms = objects.ZunService(self.context, **ms_dict)
            ms.create(self.context)
            mock_create_zun_service.assert_called_once_with(ms_dict)

    def test_destroy(self):
        with mock.patch.object(self.dbapi,
                               'get_zun_service',
                               autospec=True) as mock_get_zun_service:
            mock_get_zun_service.return_value = self.fake_zun_service
            with mock.patch.object(self.dbapi,
                                   'destroy_zun_service',
                                   autospec=True) as mock_destroy_ms:
                ms = objects.ZunService.get_by_host_and_binary(
                    self.context, 'fake-host', 'fake-bin')
                ms.destroy()
                mock_get_zun_service.assert_called_once_with(
                    'fake-host', 'fake-bin')
                mock_destroy_ms.assert_called_once_with(
                    self.fake_zun_service['host'],
                    self.fake_zun_service['binary'])
                self.assertEqual(self.context, ms._context)

    def test_save(self):
        with mock.patch.object(self.dbapi,
                               'get_zun_service',
                               autospec=True) as mock_get_zun_service:
            mock_get_zun_service.return_value = self.fake_zun_service
            with mock.patch.object(self.dbapi,
                                   'update_zun_service',
                                   autospec=True) as mock_update_ms:
                ms = objects.ZunService.get_by_host_and_binary(
                    self.context, 'fake-host', 'fake-bin')
                ms.disabled = True
                ms.save()
                mock_get_zun_service.assert_called_once_with(
                    'fake-host', 'fake-bin')
                mock_update_ms.assert_called_once_with(
                    self.fake_zun_service['host'],
                    self.fake_zun_service['binary'],
                    {'disabled': True})
                self.assertEqual(self.context, ms._context)

    def test_report_state_up(self):
        with mock.patch.object(self.dbapi,
                               'get_zun_service',
                               autospec=True) as mock_get_zun_service:
            mock_get_zun_service.return_value = self.fake_zun_service
            with mock.patch.object(self.dbapi,
                                   'update_zun_service',
                                   autospec=True) as mock_update_ms:
                ms = objects.ZunService.get_by_host_and_binary(
                    self.context, 'fake-host', 'fake-bin')
                last_report_count = self.fake_zun_service['report_count']
                ms.report_state_up()
                mock_get_zun_service.assert_called_once_with(
                    'fake-host', 'fake-bin')
                self.assertEqual(self.context, ms._context)
                mock_update_ms.assert_called_once_with(
                    self.fake_zun_service['host'],
                    self.fake_zun_service['binary'],
                    {'report_count': last_report_count + 1})

    def test_update(self):
        with mock.patch.object(self.dbapi,
                               'get_zun_service',
                               autospec=True) as mock_get_zun_service:
            mock_get_zun_service.return_value = self.fake_zun_service
            with mock.patch.object(self.dbapi,
                                   'update_zun_service',
                                   autospec=True) as mock_update_ms:
                ms = objects.ZunService.get_by_host_and_binary(
                    self.context, 'fake-host', 'fake-bin')
                kw = {'disabled': True, 'disabled_reason': 'abc'}
                ms.update(self.context, kw)
                mock_get_zun_service.assert_called_once_with(
                    'fake-host', 'fake-bin')
                mock_update_ms.assert_called_once_with(
                    self.fake_zun_service['host'],
                    self.fake_zun_service['binary'],
                    {'disabled': True,
                     'disabled_reason': 'abc'})
                self.assertEqual(self.context, ms._context)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""create volume_mapping table

Revision ID: 10d65e285a59
Revises: 945569b3669f
Create Date: 2017-08-25 22:42:18.814016

"""

# revision identifiers, used by Alembic.
revision = '10d65e285a59'
down_revision = '945569b3669f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
from sqlalchemy import dialects


def MediumText():
    return sa.Text().with_variant(dialects.mysql.MEDIUMTEXT(), 'mysql')


def upgrade():
    op.create_table(
        'volume_mapping',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(36), nullable=False),
        sa.Column('project_id', sa.String(length=255), nullable=True),
        sa.Column('user_id', sa.String(length=255), nullable=True),
        sa.Column('volume_id', sa.String(36), nullable=False),
        sa.Column('volume_provider', sa.String(36), nullable=False),
        sa.Column('container_uuid', sa.String(36), nullable=False),
        sa.Column('container_path', sa.String(length=255), nullable=True),
        sa.Column('connection_info', MediumText(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_volume0uuid'),
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import signal
import sys
import zun.conf

CONF = zun.conf.CONF

image_driver_list = [driver for driver in CONF.image_driver_list]

image_driver_list_with_none = image_driver_list + [None, 'None']

non_negative_integer = {
    'type': ['integer', 'string'],
    'pattern': '^[0-9]*$', 'minimum': 0
}

positive_integer = {
    'type': ['integer', 'string'],
    'pattern': '^[0-9]*$', 'minimum': 1
}

boolean_extended = {
    'type': ['boolean', 'string'],
    'enum': [True, 'True', 'TRUE', 'true', '1', 'ON', 'On', 'on',
             'YES', 'Yes', 'yes',
             False, 'False', 'FALSE', 'false', '0', 'OFF', 'Off', 'off',
             'NO', 'No', 'no'],
}

boolean = {
    'type': ['boolean', 'string'],
    'enum': [True, 'True', 'true', False, 'False', 'false'],
}

image_driver = {
    'type': ['string', 'null'],
    'enum': image_driver_list_with_none
}

container_name = {
    'type': ['string', 'null'],
    'minLength': 2,
    'maxLength': 255,
    'pattern': '^[a-zA-Z0-9][a-zA-Z0-9_.-]+$'
}

hex_uuid = {
    'type': 'string',
    'maxLength': 32,
    'minLength': 32,
    'pattern': '^[a-fA-F0-9]*$'
}

image_name = {
    'type': 'string',
    'minLength': 2,
    'maxLength': 255,
    'pattern': '[a-zA-Z0-9][a-zA-Z0-9_.-]'
}

image_host = {
    'type': 'string',
    'minLength': 2,
    'maxLength': 255,
    'pattern': '[a-zA-Z0-9][a-zA-Z0-9_.-]'
}

command = {
    'type': ['string', 'null']
}

command_list = {
    'type': ['array', 'null']
}

entrypoint = {
    'type': ['array', 'null']
}

auto_remove = {
    'type': ['boolean', 'null']
}

cpu = {
    'type': ['number', 'string', 'null'],
    'pattern': r'^[0-9]*(\.([0-9]+))?$',
    'minLength': 1,
    'minimum': CONF.minimum_cpus,
    'maximum': CONF.maximum_cpus,
}

cpu_policy = {
    'type': 'string',
    'enum': ['dedicated', 'shared']
}

# TODO(pksingh) Memory provided must be in MBs
# Will find another way if people dont find it useful.
memory = {
    'type': ['string', 'integer', 'null'],
    'minimum': CONF.minimum_memory,
    'maximum': CONF.maximum_memory,
    'pattern': '^[0-9]+$'
}

disk = {
    'type': ['string', 'integer', 'null'],
    'minimum': CONF.minimum_disk,
    'maximum': CONF.maximum_disk,
    'pattern': '^[0-9]+$'
}

workdir = {
    'type': ['string', 'null']
}

image_pull_policy = {
    'type': ['string', 'null'],
    'enum': ['never', 'always', 'ifnotpresent',
             'Never', 'Always', 'IfNotPresent', None]
}

labels = {
    'type': ['object', 'null']
}

annotations = {
    'type': ['object', 'null']
}

hints = {
    'type': ['object', 'null']
}

nets = {
    'type': ['array', 'null'],
    'items': {
        'type': 'object',
        'properties': {
            'network': {
                'type': ['string'],
                'minLength': 1,
                'maxLength': 255,
            },
            'v4-fixed-ip': {
                'type': ['string'],
                'format': 'ipv4'
            },
            'v6-fixed-ip': {
                'type': ['string'],
                'format': 'ipv6'
            },
            'port': {
                'type': ['string'],
                'maxLength': 255,
                'minLength': 1,
            }
        },
        'additionalProperties': False,
        'oneOf': [
            {
                'required': ['network']
            },
            {
                'required': ['port']
            }
        ]
    }
}

availability_zone = {
    'type': ['string', 'null'],
    'minLength': 1,
    'maxLength': 255,
}

healthcheck = {
    'type': ['object', 'null'],
    'items': {
        'type': 'object',
        'properties': {
            'cmd': {
                'type': ['string'],
                'minLength': 1,
                'maxLength': 255
            },
            'interval': {
                'type': ['integer', 'null']
            },
            'retries': {
                'type': ['integer', 'null']
            },
            'timeout': {
                'type': ['integer', 'null']
            }
        },
        'additionalProperties': False
    }
}

exposed_ports = {
    'type': ['object', 'null']
}

mounts = {
    'type': ['array', 'null'],
    'items': {
        'type': 'object',
        'properties': {
            'type': {
                'type': ['string'],
            },
            'source': {
                'type': ['string'],
            },
            'destination': {
                'type': ['string'],
            },
            'size': {
                'type': ['string', 'integer'],
            }
        },
        'additionalProperties': False,
        'anyOf': [
            {
                'required': ['source', 'destination']
            },
            {
                'required': ['size', 'destination']
            }
        ]
    }
}

environment = {
    'type': ['object', 'null'],
    'patternProperties': {
        '.+': {
            'type': ['string']
        },
    },
}

hostname = {
    'type': ['string', 'null'],
    'minLength': 2,
    'maxLength': 63
}

runtime = {
    'type': ['string', 'null'],
}

image_id = {
    'type': ['string', 'null'],
    'minLength': 2,
    'maxLength': 255,
    'pattern': '[a-zA-Z0-9][a-zA-Z0-9_.-]'
}

repo = {
    'type': 'string',
    'minLength': 2,
    'maxLength': 255,
    'pattern': '[a-zA-Z0-9][a-zA-Z0-9_.-]'
}


tag = copy.deepcopy(image_id)

size = {
    'type': ['string', 'integer', 'null'],
    'pattern': '^[0-9]+[b|B|k|K|m|M|g|G]?$',
}

restart_policy = {
    'type': ['object', 'null'],
    "properties": {
        "Name": {"type": ["string"],
                 "enum": ['no', 'always', 'on-failure', 'unless-stopped']},
        "MaximumRetryCount": {"type": ['integer', 'string', 'null'],
                              "minimum": 0, 'pattern': '^[0-9]+$'},
    },
    "additionalProperties": False,
    "required": ['Name']
}

string_ps_args = {
    'type': ['string'],
    'pattern': '[a-zA-Z- ,+]*'
}

str_and_int = {
    'type': ['string', 'integer', 'null'],
}

logs_since = {
    'type': ['string', 'integer', 'null'],
    'pattern': '(^[0-9]*$)|\
([0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2},[0-9]{1,3})'
}

exec_id = {
    'type': 'string',
    'maxLength': 64,
    'minLength': 64,
    'pattern': '^[a-f0-9]*$'
}

hostname = {
    'type': 'string', 'minLength': 1, 'maxLength': 255,
    # NOTE: 'host' is defined in "services" table, and that
    # means a hostname. The hostname grammar in RFC952 does
    # not allow for underscores in hostnames. However, this
    # schema allows them, because it sometimes occurs in
    # real systems.
    'pattern': '^[a-zA-Z0-9-._]*$',
}

SIGNALS = ['None']
if sys.version_info >= (3, 5, 0):
    signals = [n.name for n in signal.Signals]
    for s in signals:
        SIGNALS.append(s)
        SIGNALS.append(s.replace('SIG', ''))
        SIGNALS.append(s.lower())
        SIGNALS.append(s.lower().replace('sig', ''))
        SIGNALS.append(str(int(getattr(signal, s))))
else:
    signals = [n for n in dir(signal) if n.startswith('SIG') and '_' not in n]
    for s in signals:
        SIGNALS.append(s)
        SIGNALS.append(s.replace('SIG', ''))
        SIGNALS.append(s.lower())
        SIGNALS.append(s.lower().replace('sig', ''))
        SIGNALS.append(str(getattr(signal, s)))

signal = {
    'type': ['string', 'null'],
    'enum': SIGNALS
}

exec_command = {
    'type': ['string'],
    'minLength': 1,
}

security_groups = {
    'type': ['array', 'null'],
    'items': {
        'type': 'string',
        'minLength': 1,
        'maxLength': 255
    }
}

capsule_kind = {
    "type": ["string"],
    'enum': ['capsule', 'Capsule']
}

capsule_version = {
    "type": ["string"],
    'enum': ['beta', 'Beta']
}

capsule_metadata = {
    "type": ["object"],
    "properties": {
        "labels": labels,
        "annotations": annotations,
        # use the same format as container name
        "name": container_name,
    }
}

capsule_restart_policy = {
    "type": ["string"],
    "enum": ['Always', 'OnFailure', 'Never']
}

capsule_container_command = {
    'type': ['array'],
    'items': command
}

capsule_container_args = capsule_container_command

capsule_container_resources = {
    'type': ['object'],
    'properties': {
        'requests': {
            "type": ["object"],
            'properties': {
                'cpu': cpu,
                'memory': memory,
            },
            'additionalProperties': False,
        },
    },
    "additionalProperties": False,
}

capsule_port_protocol = {
    "type": ["string"],
    'enum': ['TCP', 'UDP']
}

capsule_container_ports = {
    'type': ['array'],
    'items': {
        'type': 'object',
        'properties': {
            'name': container_name,
            'containerPort': non_negative_integer,
            'hostPort': non_negative_integer,
            'protocol': capsule_port_protocol,
        },
        'additionalProperties': False,
        'required': ['containerPort']
    }
}

volume_name = {
    'type': ['string'],
    'minLength': 2,
    'maxLength': 255,
    'pattern': '^[a-zA-Z0-9][a-zA-Z0-9_.-]+$'
}

capsule_volume_path = {
    'type': ['string']
}

capsule_container_volume_list = {
    'type': ['array'],
    'items': {
        'type': 'object',
        'properties': {
            'name': volume_name,
            'mountPath': capsule_volume_path,
            'readOnly': boolean,
        },
        'additionalProperties': False,
        'required': ['name', 'mountPath']
    }
}

capsule_containers_list = {
    'type': ['array'],
    'items': {
        'type': 'object',
        'properties': {
            'name': container_name,
            'image': image_name,
            'command': capsule_container_command,
            'args': capsule_container_args,
            'resources': capsule_container_resources,
            'ports': capsule_container_ports,
            'volumeMounts': capsule_container_volume_list,
            'env': environment,
            'workDir': workdir,
            'imagePullPolicy': image_pull_policy,
            'stdin': boolean,
            'tty': boolean,
        },
        'additionalProperties': False,
        'required': ['image']
    }
}

volume_size = {
    'type': ['number'],
    'pattern': '^[0-9]*$',
    'minLength': 1
}

volume_auto_remove = {
    'type': boolean,
}

volume_uuid = {
    'type': 'string',
    'maxLength': 36,
    'minLength': 36
}

capsule_cinder_volume = {
    'type': 'object',
    'properties': {
        'volumeID': volume_uuid,
        'size': volume_size,
        'autoRemove': boolean,
    },
    'additionalProperties': False,
}

capsule_volumes_list = {
    'type': ['array', 'null'],
    'items': {
        'type': 'object',
        'properties': {
            'name': volume_name,
            'cinder': capsule_cinder_volume,
        },
        'additionalProperties': True,
        'required': ['name', 'cinder']
    }
}

capsule_spec = {
    'type': ['object'],
    "properties": {
        "containers": capsule_containers_list,
        "volumes": capsule_volumes_list,
        "restartPolicy": capsule_restart_policy,
    },
    "additionalProperties": True,
    "required": ['containers']
}

capsule_template = {
    'type': ['object', 'string', 'unicode'],
    "properties": {
        "kind": capsule_kind,
        "capsuleVersion": capsule_version,
        "metadata": capsule_metadata,
        # NOTE(hongbin): property 'restartPolicy' is deprecated but we keep
        # it here for backward-compatibility. Will remove it after the
        # deprecation period.
        "restartPolicy": capsule_restart_policy,
        "spec": capsule_spec,
        "availabilityZone": availability_zone,
    },
    "additionalProperties": False,
    "required": ['kind', 'spec', 'metadata']
}

neutron_net_id = {
    'type': ['string', 'null'],
    'minLength': 2,
    'maxLength': 255,
    'pattern': '[a-zA-Z0-9][a-zA-Z0-9_.-]'
}

network_name = {
    'type': ['string', 'null'],
    'minLength': 2,
    'maxLength': 255,
    'pattern': '[a-zA-Z0-9][a-zA-Z0-9_.-]'
}

registry_name = {
    'type': ['string'],
    'minLength': 2,
    'maxLength': 255,
    'pattern': '^[a-zA-Z0-9][a-zA-Z0-9_.-]+$'
}

registry_domain = {
    'type': ['string'],
    'minLength': 1,
}

registry_username = {
    'type': ['string'],
    'minLength': 1,
}

registry_password = {
    'type': ['string'],
    'minLength': 1,
}

container_registry = {
    'type': ['string'],
    'minLength': 2,
    'maxLength': 255,
    'pattern': '^[a-zA-Z0-9][a-zA-Z0-9_.-]+$'
}



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""combine tty and stdin_open

Revision ID: ce9944b346cb
Revises: 4bf34495d060
Create Date: 2017-04-21 12:24:50.201662

"""

# revision identifiers, used by Alembic.
revision = 'ce9944b346cb'
down_revision = '4bf34495d060'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    with op.batch_alter_table('container', schema=None) as batch_op:
        batch_op.add_column(sa.Column('interactive', sa.Boolean(),
                                      nullable=True))
        batch_op.drop_column('tty')
        batch_op.drop_column('stdin_open')



# Copyright 2016 Intel.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cinderclient import client as cinderclient
from glanceclient import client as glanceclient
from keystoneauth1.loading import adapter as ka_adapter
from neutronclient.v2_0 import client as neutronclient

from zun.common import exception
from zun.common import keystone
import zun.conf


class OpenStackClients(object):
    """Convenience class to create and cache client instances."""

    def __init__(self, context):
        self.context = context
        self._keystone = None
        self._glance = None
        self._neutron = None
        self._cinder = None
        self._placement = None
        self._placement_ks_filter = None

    def url_for(self, **kwargs):
        return self.keystone().session.get_endpoint(**kwargs)

    def zun_url(self):
        endpoint_type = self._get_client_option('zun', 'endpoint_type')
        region_name = self._get_client_option('zun', 'region_name')
        return self.url_for(service_type='container',
                            interface=endpoint_type,
                            region_name=region_name)

    @property
    def auth_token(self):
        return self.context.auth_token or self.keystone().auth_token

    def keystone(self):
        if self._keystone:
            return self._keystone

        self._keystone = keystone.KeystoneClientV3(self.context)
        return self._keystone

    def _get_client_option(self, client, option):
        return getattr(getattr(zun.conf.CONF, '%s_client' % client), option)

    @exception.wrap_keystone_exception
    def glance(self):
        if self._glance:
            return self._glance

        glanceclient_version = self._get_client_option('glance', 'api_version')
        session = self.keystone().session

        endpoint_type = self._get_client_option('glance', 'endpoint_type')
        region_name = self._get_client_option('glance', 'region_name')
        glance_url = self.url_for(service_type='image',
                                  interface=endpoint_type,
                                  region_name=region_name)

        self._glance = glanceclient.Client(glanceclient_version,
                                           endpoint=glance_url,
                                           session=session)

        return self._glance

    @exception.wrap_keystone_exception
    def neutron(self):
        if self._neutron:
            return self._neutron

        session = self.keystone().session
        if self._get_client_option('neutron', 'ca_file'):
            session.verify = self._get_client_option('neutron', 'ca_file')
        if self._get_client_option('neutron', 'insecure'):
            session.verify = False
        endpoint_type = self._get_client_option('neutron', 'endpoint_type')
        region_name = self._get_client_option('neutron', 'region_name')
        self._neutron = neutronclient.Client(session=session,
                                             endpoint_type=endpoint_type,
                                             region_name=region_name)

        return self._neutron

    @exception.wrap_keystone_exception
    def cinder(self):
        if self._cinder:
            return self._cinder

        session = self.keystone().session
        if self._get_client_option('cinder', 'ca_file'):
            session.verify = self._get_client_option('cinder', 'ca_file')
        if self._get_client_option('cinder', 'insecure'):
            session.verify = False
        cinder_api_version = self._get_client_option('cinder', 'api_version')
        region_name = self._get_client_option('cinder', 'region_name')
        endpoint_type = self._get_client_option('cinder', 'endpoint_type')
        kwargs = {
            'session': self.keystone().session,
            'endpoint_type': endpoint_type,
            'region_name': region_name,
            'cacert': self._get_client_option('cinder', 'ca_file'),
            'insecure': self._get_client_option('cinder', 'insecure')
        }
        self._cinder = cinderclient.Client(version=cinder_api_version,
                                           **kwargs)

        return self._cinder

    @exception.wrap_keystone_exception
    def placement(self):
        if self._placement:
            return self._placement, self._placement_ks_filter

        session = self.keystone().session
        if self._get_client_option('placement', 'ca_file'):
            session.verify = self._get_client_option('placement', 'ca_file')
        if self._get_client_option('placement', 'insecure'):
            session.verify = False
        region_name = self._get_client_option('placement', 'region_name')
        endpoint_type = self._get_client_option('placement', 'endpoint_type')
        kwargs = {
            'session': self.keystone().session,
            'auth': self.keystone().auth,
        }
        self._placement_ks_filter = {'service_type': 'placement',
                                     'region_name': region_name,
                                     'interface': endpoint_type}
        self._placement = ka_adapter.Adapter().load_from_options(**kwargs)

        return self._placement, self._placement_ks_filter



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""add host to volume mapping

Revision ID: f746cd28bcac
Revises: 2fb377a5a519
Create Date: 2018-08-03 10:53:45.920787

"""

# revision identifiers, used by Alembic.
revision = 'f746cd28bcac'
down_revision = '2fb377a5a519'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('volume_mapping',
                  sa.Column('host', sa.String(length=255), nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""remove meta from container

Revision ID: c2052ead4f95
Revises: df87dbd4846c
Create Date: 2019-12-30 17:11:48.838977

"""

# revision identifiers, used by Alembic.
revision = 'c2052ead4f95'
down_revision = 'df87dbd4846c'
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    op.drop_column('container', 'meta')



# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shutil
import sys

from oslo_upgradecheck import common_checks
from oslo_upgradecheck import upgradecheck

from zun.common.i18n import _
import zun.conf

CONF = zun.conf.CONF


class Checks(upgradecheck.UpgradeCommands):

    """Contains upgrade checks

    Various upgrade checks should be added as separate methods in this class
    and added to _upgrade_checks tuple.
    """

    def _cmd_exists(self, cmd):
        try:
            return shutil.which(cmd) is not None
        except AttributeError:
            # shutil.which is not available in python 2.x so try an
            # alternative approach
            return any(
                os.access(os.path.join(path, cmd), os.X_OK)
                for path in os.environ["PATH"].split(os.pathsep)
            )

    def _numactl_check(self):
        """This is a check for existence of numactl binary

        It needs to be removed after adding any real upgrade check
        """
        if self._cmd_exists('numactl'):
            return upgradecheck.Result(upgradecheck.Code.SUCCESS)
        else:
            msg = _("The program 'numactl' is currently not installed.")
            return upgradecheck.Result(upgradecheck.Code.FAILURE, msg)

    _upgrade_checks = (
        (_('Numactl Check'), _numactl_check),
        (_('Policy File JSON to YAML Migration'),
         (common_checks.check_policy_json, {'conf': CONF})),
    )


def main():
    return upgradecheck.main(
        CONF, project='zun', upgrade_command=Checks())


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from zun.common import exception
from zun import objects
from zun.objects import fields
from zun.pci import stats
from zun.pci import whitelist
from zun.tests import base
from zun.tests.unit.pci import fakes

CONF = cfg.CONF
fake_pci_1 = {
    'compute_node_uuid': 1,
    'address': '0000:00:00.1',
    'product_id': 'p1',
    'vendor_id': 'v1',
    'status': 'available',
    'extra_k1': 'v1',
    'request_id': None,
    'numa_node': 0,
    'dev_type': fields.PciDeviceType.STANDARD,
    'parent_addr': None,
    }


fake_pci_2 = dict(fake_pci_1, vendor_id='v2',
                  product_id='p2',
                  address='0000:00:00.2',
                  numa_node=1)


fake_pci_3 = dict(fake_pci_1, address='0000:00:00.3')

fake_pci_4 = dict(fake_pci_1, vendor_id='v3',
                  product_id='p3',
                  address='0000:00:00.3',
                  numa_node=None)


class PciDeviceStatsTestCase(base.TestCase):
    def _create_fake_devs(self):
        self.fake_dev_1 = objects.PciDevice.create(None, fake_pci_1)
        self.fake_dev_2 = objects.PciDevice.create(None, fake_pci_2)
        self.fake_dev_3 = objects.PciDevice.create(None, fake_pci_3)
        self.fake_dev_4 = objects.PciDevice.create(None, fake_pci_4)

        for dev in [self.fake_dev_1, self.fake_dev_2,
                    self.fake_dev_3, self.fake_dev_4]:
            self.pci_stats.add_device(dev)

    def setUp(self):
        super(PciDeviceStatsTestCase, self).setUp()
        self.pci_stats = stats.PciDeviceStats()
        # The following two calls need to be made before adding the devices.
        patcher = fakes.fake_pci_whitelist()
        self.addCleanup(patcher.stop)
        self._create_fake_devs()

    def test_add_device(self):
        self.assertEqual(len(self.pci_stats.pools), 3)
        self.assertEqual(set([d['vendor_id'] for d in self.pci_stats]),
                         set(['v1', 'v2', 'v3']))
        self.assertEqual(set([d['count'] for d in self.pci_stats]),
                         set([1, 2]))

    def test_remove_device(self):
        self.pci_stats.remove_device(self.fake_dev_2)
        self.assertEqual(len(self.pci_stats.pools), 2)
        self.assertEqual(self.pci_stats.pools[0]['count'], 2)
        self.assertEqual(self.pci_stats.pools[0]['vendor_id'], 'v1')

    def test_remove_device_exception(self):
        self.pci_stats.remove_device(self.fake_dev_2)
        self.assertRaises(exception.PciDevicePoolEmpty,
                          self.pci_stats.remove_device,
                          self.fake_dev_2)

    def test_pci_stats_equivalent(self):
        pci_stats2 = stats.PciDeviceStats()
        for dev in [self.fake_dev_1,
                    self.fake_dev_2,
                    self.fake_dev_3,
                    self.fake_dev_4]:
            pci_stats2.add_device(dev)
        self.assertEqual(self.pci_stats, pci_stats2)

    def test_pci_stats_not_equivalent(self):
        pci_stats2 = stats.PciDeviceStats()
        for dev in [self.fake_dev_1,
                    self.fake_dev_2,
                    self.fake_dev_3]:
            pci_stats2.add_device(dev)
        self.assertNotEqual(self.pci_stats, pci_stats2)

    def test_object_create(self):
        m = self.pci_stats.to_device_pools_obj()
        new_stats = stats.PciDeviceStats(m)

        self.assertEqual(len(new_stats.pools), 3)
        self.assertEqual(set([d['count'] for d in new_stats]),
                         set([1, 2]))
        self.assertEqual(set([d['vendor_id'] for d in new_stats]),
                         set(['v1', 'v2', 'v3']))

    @mock.patch(
        'zun.pci.whitelist.Whitelist._parse_white_list_from_config')
    def test_white_list_parsing(self, mock_whitelist_parse):
        white_list = '{"product_id":"0001", "vendor_id":"8086"}'
        CONF.set_override('passthrough_whitelist', white_list, 'pci')
        pci_stats = stats.PciDeviceStats()
        pci_stats.add_device(self.fake_dev_2)
        pci_stats.remove_device(self.fake_dev_2)
        self.assertEqual(1, mock_whitelist_parse.call_count)


class PciDeviceStatsWithTagsTestCase(base.TestCase):

    def setUp(self):
        super(PciDeviceStatsWithTagsTestCase, self).setUp()
        white_list = ['{"vendor_id":"1137","product_id":"0071",'
                      '"address":"*:0a:00.*","physical_network":"physnet1"}',
                      '{"vendor_id":"1137","product_id":"0072"}']
        self.config(passthrough_whitelist=white_list, group='pci')
        dev_filter = whitelist.Whitelist(white_list)
        self.pci_stats = stats.PciDeviceStats(dev_filter=dev_filter)

    def _create_pci_devices(self):
        self.pci_tagged_devices = []
        for dev in range(4):
            pci_dev = {'compute_node_uuid': 1,
                       'address': '0000:0a:00.%d' % dev,
                       'vendor_id': '1137',
                       'product_id': '0071',
                       'status': 'available',
                       'request_id': None,
                       'dev_type': 'PCI',
                       'parent_addr': None,
                       'numa_node': 0}
            self.pci_tagged_devices.append(objects.PciDevice.create(None,
                                                                    pci_dev))

        self.pci_untagged_devices = []
        for dev in range(3):
            pci_dev = {'compute_node_uuid': 1,
                       'address': '0000:0b:00.%d' % dev,
                       'vendor_id': '1137',
                       'product_id': '0072',
                       'status': 'available',
                       'request_id': None,
                       'dev_type': 'PCI',
                       'parent_addr': None,
                       'numa_node': 0}
            self.pci_untagged_devices.append(objects.PciDevice.create(None,
                                                                      pci_dev))

        for dev in self.pci_tagged_devices:
            self.pci_stats.add_device(dev)

        for dev in self.pci_untagged_devices:
            self.pci_stats.add_device(dev)

    def _assertPoolContent(self, pool, vendor_id, product_id, count, **tags):
        self.assertEqual(vendor_id, pool['vendor_id'])
        self.assertEqual(product_id, pool['product_id'])
        self.assertEqual(count, pool['count'])
        if tags:
            for k, v in tags.items():
                self.assertEqual(v, pool[k])

    def _assertPools(self):
        # Pools are ordered based on the number of keys. 'product_id',
        # 'vendor_id' are always part of the keys. When tags are present,
        # they are also part of the keys. In this test class, we have
        # two pools with the second one having the tag 'physical_network'
        # and the value 'physnet1'
        self.assertEqual(2, len(self.pci_stats.pools))
        self._assertPoolContent(self.pci_stats.pools[0], '1137', '0072',
                                len(self.pci_untagged_devices))
        self.assertEqual(self.pci_untagged_devices,
                         self.pci_stats.pools[0]['devices'])
        self._assertPoolContent(self.pci_stats.pools[1], '1137', '0071',
                                len(self.pci_tagged_devices),
                                physical_network='physnet1')
        self.assertEqual(self.pci_tagged_devices,
                         self.pci_stats.pools[1]['devices'])

    def test_add_devices(self):
        self._create_pci_devices()
        self._assertPools()

    def test_add_device_no_devspec(self):
        self._create_pci_devices()
        pci_dev = {'compute_node_uuid': 1,
                   'address': '0000:0c:00.1',
                   'vendor_id': '2345',
                   'product_id': '0172',
                   'status': 'available',
                   'parent_addr': None,
                   'request_id': None}
        pci_dev_obj = objects.PciDevice.create(None, pci_dev)
        self.pci_stats.add_device(pci_dev_obj)
        # There should be no change
        self.assertIsNone(
            self.pci_stats._create_pool_keys_from_dev(pci_dev_obj))
        self._assertPools()

    def test_remove_device_no_devspec(self):
        self._create_pci_devices()
        pci_dev = {'compute_node_uuid': 1,
                   'address': '0000:0c:00.1',
                   'vendor_id': '2345',
                   'product_id': '0172',
                   'status': 'available',
                   'parent_addr': None,
                   'request_id': None}
        pci_dev_obj = objects.PciDevice.create(None, pci_dev)
        self.pci_stats.remove_device(pci_dev_obj)
        # There should be no change
        self.assertIsNone(
            self.pci_stats._create_pool_keys_from_dev(pci_dev_obj))
        self._assertPools()

    def test_remove_device(self):
        self._create_pci_devices()
        dev1 = self.pci_untagged_devices.pop()
        self.pci_stats.remove_device(dev1)
        dev2 = self.pci_tagged_devices.pop()
        self.pci_stats.remove_device(dev2)
        self._assertPools()


class PciDeviceVFPFStatsTestCase(base.TestCase):

    def setUp(self):
        super(PciDeviceVFPFStatsTestCase, self).setUp()
        white_list = ['{"vendor_id":"8086","product_id":"1528"}',
                      '{"vendor_id":"8086","product_id":"1515"}']
        self.config(passthrough_whitelist=white_list, group='pci')
        self.pci_stats = stats.PciDeviceStats()

    def _create_pci_devices(self, vf_product_id=1515, pf_product_id=1528):
        self.sriov_pf_devices = []
        for dev in range(2):
            pci_dev = {'compute_node_uuid': 1,
                       'address': '0000:81:00.%d' % dev,
                       'vendor_id': '8086',
                       'product_id': '%d' % pf_product_id,
                       'status': 'available',
                       'request_id': None,
                       'dev_type': fields.PciDeviceType.SRIOV_PF,
                       'parent_addr': None,
                       'numa_node': 0}
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.child_devices = []
            self.sriov_pf_devices.append(dev_obj)

        self.sriov_vf_devices = []
        for dev in range(8):
            pci_dev = {'compute_node_uuid': 1,
                       'address': '0000:81:10.%d' % dev,
                       'vendor_id': '8086',
                       'product_id': '%d' % vf_product_id,
                       'status': 'available',
                       'request_id': None,
                       'dev_type': fields.PciDeviceType.SRIOV_VF,
                       'parent_addr': '0000:81:00.%d' % int(dev / 4),
                       'numa_node': 0}
            dev_obj = objects.PciDevice.create(None, pci_dev)
            dev_obj.parent_device = self.sriov_pf_devices[int(dev / 4)]
            dev_obj.parent_device.child_devices.append(dev_obj)
            self.sriov_vf_devices.append(dev_obj)

        list(map(self.pci_stats.add_device, self.sriov_pf_devices))
        list(map(self.pci_stats.add_device, self.sriov_vf_devices))



# Copyright 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Define supported virtual NIC types. VNIC_TYPE_DIRECT and VNIC_TYPE_MACVTAP
# are used for SR-IOV ports
VNIC_TYPE_NORMAL = 'normal'
VNIC_TYPE_DIRECT = 'direct'
VNIC_TYPE_MACVTAP = 'macvtap'
VNIC_TYPE_DIRECT_PHYSICAL = 'direct-physical'
VNIC_TYPE_BAREMETAL = 'baremetal'
VNIC_TYPE_VIRTIO_FORWARDER = 'virtio-forwarder'

# Define list of ports which needs pci request.
# Note: The macvtap port needs a PCI request as it is a tap interface
# with VF as the lower physical interface.
# Note: Currently, VNIC_TYPE_VIRTIO_FORWARDER assumes a 1:1
# relationship with a VF. This is expected to change in the future.
VNIC_TYPES_SRIOV = (VNIC_TYPE_DIRECT, VNIC_TYPE_MACVTAP,
                    VNIC_TYPE_DIRECT_PHYSICAL, VNIC_TYPE_VIRTIO_FORWARDER)

# Define list of ports which are passthrough to the guest
# and need a special treatment on snapshot and suspend/resume
VNIC_TYPES_DIRECT_PASSTHROUGH = (VNIC_TYPE_DIRECT,
                                 VNIC_TYPE_DIRECT_PHYSICAL)
